import tensorflow as tf
from arenets.context.configurations.rcnn import RCNNConfig
from arenets.tf_helpers.cell_types import CellTypes


class AttentionSelfPZhouRCNNConfig(RCNNConfig):

    __cell_type = CellTypes.LSTM

    def __init__(self):
        super(AttentionSelfPZhouRCNNConfig, self).__init__()

    # region properties

    @property
    def CellType(self):
        return self.__cell_type

    @property
    def BiasInitializer(self):
        return tf.constant_initializer(0.1)

    @property
    def WeightInitializer(self):
        return tf.contrib.layers.xavier_initializer()

    # endregion


import tensorflow as tf
from collections import OrderedDict

from arenets.context.architectures.base.fc_single import FullyConnectedLayer
from arenets.context.configurations.rcnn import RCNNConfig
from arenets.sample import InputSample
from arenets.arekit.common.data_type import DataType
from arenets.tf_helpers import sequence


class RCNN(FullyConnectedLayer):
    """
    Copyright (c) Joohong Lee
    Title: Recurrent Convolutional Neural Networks for Text Classification
    Paper: https://www.aaai.org/ocs/index.php/AAAI/AAAI15/paper/view/9745
    Source: https://github.com/roomylee/rcnn-text-classification
    """

    H_W_text = "W_text"
    H_b_text = "b_text"

    def __init__(self):
        super(RCNN, self).__init__()
        self.__hidden = OrderedDict()
        self.__dropout_rnn_keep_prob = None

    # region properties

    @property
    def ContextEmbeddingSize(self):
        return self.Config.HiddenSize

    # endregion

    # region public 'set' methods

    def set_input_rnn_keep_prob(self, value):
        self.__dropout_rnn_keep_prob = value

    # endregion

    # region public 'init' methods

    def init_input(self):
        super(RCNN, self).init_input()
        self.__dropout_rnn_keep_prob = tf.compat.v1.placeholder(dtype=tf.float32,
                                                                name="ctx_dropout_rnn_keep_prob")

    def modify_rnn_outputs_optional(self, output_fw, output_bw):
        # Nothing modifies
        return output_fw, output_bw

    def init_context_embedding(self, embedded_terms):
        assert(isinstance(self.Config, RCNNConfig))
        text_length = sequence.calculate_sequence_length(self.get_input_parameter(InputSample.I_X_INDS))

        with tf.name_scope("bi-rnn"):

            fw_cell = sequence.get_cell(hidden_size=self.Config.SurroundingOneSideContextEmbeddingSize,
                                        cell_type=self.Config.CellType,
                                        dropout_rnn_keep_prob=self.__dropout_rnn_keep_prob)

            bw_cell = sequence.get_cell(hidden_size=self.Config.SurroundingOneSideContextEmbeddingSize,
                                        cell_type=self.Config.CellType,
                                        dropout_rnn_keep_prob=self.__dropout_rnn_keep_prob)

            (output_fw, output_bw), states = sequence.bidirectional_rnn(
                cell_fw=fw_cell,
                cell_bw=bw_cell,
                inputs=embedded_terms,
                sequence_length=text_length,
                dtype=tf.float32)

            output_fw, output_bw = self.modify_rnn_outputs_optional(output_fw, output_bw)

        with tf.name_scope("ctx"):
            shape = [tf.shape(output_fw)[0], 1, tf.shape(output_fw)[2]]
            c_left = tf.concat([tf.zeros(shape), output_fw[:, :-1]], axis=1, name="context_left")
            c_right = tf.concat([output_bw[:, 1:], tf.zeros(shape)], axis=1, name="context_right")

        with tf.name_scope("word-representation"):
            merged = tf.concat([c_left, embedded_terms, c_right], axis=2, name="merged")

        with tf.name_scope("text-representation"):
            y2 = tf.tanh(tf.einsum('aij,jk->aik', merged, self.__hidden[self.H_W_text]) + self.__hidden[self.H_b_text])

        with tf.name_scope("max-pooling"):
            y3 = tf.reduce_max(y2, axis=1)

        return y3

    def init_body_dependent_hidden_states(self):
        assert(isinstance(self.Config, RCNNConfig))

        self.__hidden[self.H_W_text] = tf.compat.v1.get_variable(
            name=self.H_W_text,
            shape=[self.__text_embedding_size(), self.Config.HiddenSize],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.WeightInitializer)

        self.__hidden[self.H_b_text] = tf.compat.v1.get_variable(
            name=self.H_b_text,
            shape=[self.Config.HiddenSize],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.BiasInitializer)

    # endregion

    # region public 'iter' methods

    def iter_hidden_parameters(self):
        for key, value in super(RCNN, self).iter_hidden_parameters():
            yield key, value

        for key, value in self.__hidden.items():
            yield key, value

    # endregion

    # region public 'create' methods

    def create_feed_dict(self, input, data_type):
        feed_dict = super(RCNN, self).create_feed_dict(input=input, data_type=data_type)
        feed_dict[self.__dropout_rnn_keep_prob] = self.Config.DropoutRNNKeepProb if data_type == DataType.Train else 1.0
        return feed_dict

    # endregion

    # region private methods

    def __text_embedding_size(self):
        return self.TermEmbeddingSize + \
               2 * self.Config.SurroundingOneSideContextEmbeddingSize

    # endregion



import tensorflow as tf

from arenets.attention.configurations.mlp import MLPAttentionConfig
from arenets.tf_helpers.layers import get_k_layer_logits
from arenets.tf_helpers.filtering import \
    filter_batch_elements, \
    select_entity_related_elements


class MLPAttention(object):
    """
    Title: Attention-Based Convolutional Neural Network for Semantic Relation Extraction
    Authors: Yatian Shen, Xuanjing Huang
    Paper: https://www.aclweb.org/anthology/C16-1238
    """

    H_W_we = "H_W_we"
    H_W_a = "H_W_a"
    H_b_we = "H_b_we"
    H_b_a = "H_b_a"

    I_keys = "I_keys"

    def __init__(self, cfg, batch_size, terms_per_context):
        assert(isinstance(cfg, MLPAttentionConfig))
        self.__cfg = cfg

        self.__batch_size = batch_size
        self.__terms_per_context = terms_per_context
        self.__term_embedding_size = None

        self.__input = {}
        self.__hidden = {}

    # region properties

    @property
    def Config(self):
        return self.__cfg

    @property
    def BatchSize(self):
        return self.__batch_size

    @property
    def TermsPerContext(self):
        return self.__terms_per_context

    @property
    def TermEmbeddingSize(self):
        return self.__term_embedding_size

    @property
    def Input(self):
        return self.__input

    @property
    def AttentionEmbeddingSize(self):
        return self.__cfg.KeysPerContext * self.__term_embedding_size

    # endregion

    # region public methods

    def set_input(self, param_names_with_values, keys):
        """
        param_names_with_values: list
            list of pairs <name, input>
        """
        assert(isinstance(param_names_with_values, list))

        self.__input[self.I_keys] = keys

        for p_name, value in param_names_with_values:
            self.__input[p_name] = value

    def init_hidden(self):

        self.__hidden[self.H_W_we] = tf.compat.v1.get_variable(
            name=self.H_W_we,
            shape=[2 * self.__term_embedding_size, self.__cfg.HiddenSize],
            initializer=self.__cfg.LayerInitializer,
            dtype=tf.float32)

        self.__hidden[self.H_b_we] = tf.compat.v1.get_variable(
            name=self.H_b_we,
            shape=[self.__cfg.HiddenSize],
            initializer=self.__cfg.LayerInitializer,
            dtype=tf.float32)

        self.__hidden[self.H_W_a] = tf.compat.v1.get_variable(
            name=self.H_W_a,
            shape=[self.__cfg.HiddenSize, 1],
            initializer=self.__cfg.LayerInitializer,
            dtype=tf.float32)

        self.__hidden[self.H_b_a] = tf.compat.v1.get_variable(
            name=self.H_b_a,
            shape=[1],
            initializer=self.__cfg.LayerInitializer,
            dtype=tf.float32)

    def init_term_embedding_size(self, p_names_with_sizes):
        assert(isinstance(p_names_with_sizes, list))
        self.__term_embedding_size = sum([size for _, size in p_names_with_sizes])

    def init_body(self, params_embeddings):
        """
        params_embedding: list
            list of pairs <name, embedding>
        """
        assert(isinstance(params_embeddings, list))

        embedded_terms = tf.concat(
            values=[tf.nn.embedding_lookup(params=p_emb, ids=self.__input[p_name])
                    for p_name, p_emb in params_embeddings],
            axis=-1)

        with tf.name_scope("attention"):

            def iter_by_entities(entities, handler):
                """
                entities:  [batch_size, entities_per_context]
                handler: func
                """

                e_len = self.calculate_keys_length_func(entities)

                att_sum_array = tf.TensorArray(
                    dtype=tf.float32,
                    name="context_iter",
                    size=e_len,
                    infer_shape=False,
                    dynamic_size=True)

                att_weights_array = tf.TensorArray(
                    dtype=tf.float32,
                    name="context_iter",
                    size=e_len,
                    infer_shape=False,
                    dynamic_size=True)

                _, _, att_sum, att_weights = tf.while_loop(
                    lambda i, *_: tf.less(i, e_len),
                    handler,
                    (0, entities, att_sum_array, att_weights_array))

                return att_sum.stack(), att_weights.stack()

            def process_entity(i, entities, att_sum, att_weights):
                """
                entities: [batch_size, entities_per_context]
                params_with_embedding: list
                    list of pairs <input, embedding>
                """
                e_term_index = tf.gather(entities, [i], axis=1)                         # [batch_size, 1] -- term positions
                e_term_indices = tf.tile(e_term_index, [1, self.__terms_per_context])   # [batch_size, terms_per_context]

                embedded_params = []

                # p_name: [batch_size, terms_per_context]

                for param_name, param_embedding in params_embeddings:
                    ids = filter_batch_elements(elements=self.__input[param_name],
                                                inds=e_term_indices,
                                                handler=select_entity_related_elements)
                    embedded_params.append(tf.nn.embedding_lookup(params=param_embedding, ids=ids))

                e = tf.concat(embedded_params, axis=-1)  # [batch_size, terms_per_context, embedding_size]

                merged = tf.concat([embedded_terms, e], axis=-1)
                merged = tf.reshape(merged, [self.__batch_size * self.__terms_per_context,
                                             2 * self.__term_embedding_size])

                u = get_k_layer_logits(g=merged,
                                       W=[self.__hidden[self.H_W_we], self.__hidden[self.H_W_a]],
                                       b=[self.__hidden[self.H_b_we], self.__hidden[self.H_b_a]],
                                       activations=[None,
                                                    lambda tensor: tf.tanh(tensor),
                                                    None])       # [batch_size * terms_per_context, 1]

                alphas = tf.reshape(u, [self.__batch_size, self.__terms_per_context])
                alphas = tf.nn.softmax(alphas)
                alphas = tf.reshape(alphas, [self.__batch_size * self.__terms_per_context, 1])

                original_embedding = tf.reshape(embedded_terms,
                                                [self.__batch_size * self.__terms_per_context, self.__term_embedding_size])

                w_embedding = tf.multiply(alphas, original_embedding)
                w_embedding = tf.reshape(w_embedding, [self.__batch_size, self.__terms_per_context, self.__term_embedding_size])
                w_sum = tf.reduce_sum(w_embedding, axis=1)             # [batch_size, embedding_size]

                return (i + 1,
                        entities,
                        att_sum.write(i, w_sum),
                        att_weights.write(i, tf.reshape(alphas, [self.__batch_size, self.__terms_per_context])))

            # I_keys: [batch_size, EntitiesPerContext]
            att_sum, att_weights = iter_by_entities(
                entities=self.__input[self.I_keys],
                handler=process_entity)

            # att_weights: [entity_per_context, batch_size, terms_per_context]
            # att_sum: [entity_per_context, batch_size, term_embedding_size]
            att_weights = tf.transpose(att_weights, perm=[1, 0, 2])  # [batch_size, entity_per_context, terms_per_context]
            att_sum = tf.transpose(att_sum, perm=[1, 0, 2])  # [batch_size, entity_per_context, term_embedding_size]

        return self.reshape_att_sum(att_sum), \
               self.reshape_att_weights(att_weights)

    def reshape_att_sum(self, att_sum):
        """
        att_sum: [batch_size, entity_per_context, term_embedding_size]
        """
        att_sum = tf.reshape(att_sum, shape=[self.__batch_size, self.AttentionEmbeddingSize])
        return att_sum

    def reshape_att_weights(self, att_weights):
        """
        att_sum: [batch_size, entity_per_context, term_embedding_size]
        """
        return att_weights

    def calculate_keys_length_func(self, keys):
        """
        In this case we consider that the length is fixed to EntitiesPerContextParameter
        """
        return self.__cfg.KeysPerContext

    # endregion



from os.path import dirname, join

import gzip
import sys
import unittest

sys.path.append('../')

from arenets.core.input.rows_parser import ParsedSampleRow
from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.sample import InputSample
from arenets.shapes import NetworkInputShapes
from arenets.external.readers.pandas_csv_reader import PandasCsvReader


class TestSamplesIteration(unittest.TestCase):

    __show_examples = False
    __show_shifted_examples = False

    def __get_local_dir(self, local_filepath):
        return join(dirname(__file__), local_filepath)

    def test_check_all_samples(self):
        vocab_filepath = self.__get_local_dir("test_data/vocab.txt.gz")
        samples_filepath = self.__get_local_dir("test_data/sample-train.tsv.gz")
        terms_vocab = self.__read_vocab(vocab_filepath)
        config = DefaultNetworkConfig()
        config.modify_terms_per_context(50)

        self.__test_core(terms_vocab=terms_vocab,
                         config=config,
                         samples_filepath=samples_filepath)

    def test_show_all_samples(self):
        self.__show_examples = True
        self.test_check_all_samples()

    def test_show_shifted_examples_only(self):
        self.__show_examples = True
        self.__show_shifted_examples = True
        self.test_check_all_samples()

    # region private methods

    @staticmethod
    def __iter_tsv_gzip(input_file):
        """Reads a tab separated value file."""
        reader = PandasCsvReader(compression='gzip', sep='\t', encoding='utf-8')
        for _, row in reader.read(input_file):
            yield row

    @staticmethod
    def __read_vocab(input_file):
        words = {}
        with gzip.open(input_file, mode="rt") as f:
            for w_ind, line in enumerate(f.readlines()):
                w = line.strip()
                if w in words:
                    raise Exception("Word already presented: {}".format(w).encode('utf-8'))
                words[w] = w_ind
        return words

    def __terms_to_text_line(self, terms, frame_inds_set):
        assert(isinstance(frame_inds_set, set))
        words = []
        for t_index, t_value in enumerate(terms):
            value = t_value
            if t_index in frame_inds_set:
                value = "[[{}]]".format(t_value)
            words.append(value)
        return " ".join(words)

    def __test_core(self, terms_vocab, config, samples_filepath):
        assert(isinstance(config, DefaultNetworkConfig))
        assert(isinstance(samples_filepath, str))

        samples = []
        for i, row in enumerate(self.__iter_tsv_gzip(input_file=samples_filepath)):

            # Perform row parsing process.
            row = ParsedSampleRow(row)

            subj_ind = row.SubjectIndex
            obj_ind = row.ObjectIndex
            input_shapes = NetworkInputShapes(
                [(NetworkInputShapes.SYNONYMS_PER_CONTEXT, config.TermsPerContext),
                 (NetworkInputShapes.TERMS_PER_CONTEXT, config.TermsPerContext),
                 (NetworkInputShapes.FRAMES_PER_CONTEXT, config.FramesPerContext)])

            sample = InputSample.create_from_parameters(
                input_sample_id=row.SampleID,
                terms=row.Terms,
                entity_inds=row.EntityInds,
                subj_ind=int(row.SubjectIndex),
                obj_ind=int(row.ObjectIndex),
                terms_vocab=terms_vocab,
                is_external_vocab=True,
                input_shapes=input_shapes,
                frame_inds=row.TextFrameVariantIndices,
                frame_connotations=row.TextFrameConnotations,
                pos_tags=row.PartOfSpeechTags,
                syn_subj_inds=row.SynonymSubjectInds,
                syn_obj_inds=row.SynonymObjectInds,
                unknown_term_index=0)

            if sample._shift_index_dbg == 0 and self.__show_shifted_examples:
                continue

            if self.__show_examples:
                print("------------------")
                print("INPUT SAMPLE DATA")
                print("------------------")
                print("offset index (debug): {}".format(sample._shift_index_dbg))
                print("id: {}".format(row.SampleID))
                print("label: {}".format(row.UintLabel))
                print("entity_inds: {}".format(row.EntityInds))
                print("subj_ind: {}".format(subj_ind))
                print("obj_ind: {}".format(obj_ind))
                print("frame_inds: {}".format(row.TextFrameVariantIndices))
                print("frame_connots_uint: {}".format(row.TextFrameConnotations))
                print("syn_obj: {}".format(row.SynonymObjectInds))
                print("syn_subj: {}".format(row.SynonymSubjectInds))
                print("terms: {}".format(row.Terms))
                print("pos_tags: {}".format(row.PartOfSpeechTags))

                print(self.__terms_to_text_line(terms=row.Terms, frame_inds_set=set(row.TextFrameVariantIndices)))

                print("------------------")
                print("NETWORK INPUT DATA")
                print("------------------")
                for key, value in sample:
                    print("{key}:\n{value}".format(key=key, value=value))

            samples.append(sample)

    # endregion


if __name__ == '__main__':
    unittest.main()



from arenets.arekit.common.data_type import DataType
from arenets.core.feeding.batch.base import MiniBatch
from arenets.tf_helpers.session import initialize_session


class TensorflowModelContext(object):

    FeedDictShow = False

    def __init__(self, network, config, nn_io, bags_collection_type, inference_ctx):

        self.__sess = None
        self.__optimiser = None
        self.__inference_ctx = inference_ctx
        self.__network = network
        self.__io = nn_io

        self.__config = config
        self.__bags_collection_type = bags_collection_type

    # region property

    @property
    def IO(self):
        return self.__io

    @property
    def Config(self):
        return self.__config

    @property
    def Network(self):
        return self.__network

    @property
    def Session(self):
        return self.__sess

    @property
    def Optimiser(self):
        return self.__optimiser

    @property
    def BagsCollectionType(self):
        return self.__bags_collection_type

    # endregion

    def __set_optimiser_value(self, value):
        self.__optimiser = value

    def get_bags_collection(self, data_type):
        return self.__inference_ctx.BagsCollections[data_type]

    def get_sample_id_label_pairs(self, data_type):
        return self.__inference_ctx.SampleIdAndLabelPairs[data_type]

    def set_optimiser(self):
        optimiser = self.Config.Optimiser.minimize(self.__network.Cost)
        self.__set_optimiser_value(optimiser)

    def initialize_session(self):

        if self.__sess is not None:
            return

        self.__sess = initialize_session()

    def dispose_session(self):
        """ Tensorflow session dispose method
        """
        self.__sess.close()

    def create_feed_dict(self, minibatch, data_type):
        """ Compose feeding dictionary from minibatch.
        """
        assert(isinstance(minibatch, MiniBatch))
        assert(isinstance(data_type, DataType))

        network_input = minibatch.to_network_input(provide_labels=data_type != DataType.Test)

        if self.FeedDictShow:
            MiniBatch.debug_output(network_input)

        return self.__network.create_feed_dict(input=network_input, data_type=data_type)



from arenets.features.utils import create_zeros


def calculate_term_types(terms, entity_inds_set):
    assert(isinstance(terms, list))
    assert(isinstance(entity_inds_set, set))

    vector = create_zeros(size=len(terms))

    for t_index, term in enumerate(terms):
        if t_index in entity_inds_set:
            vector[t_index] = 1

    return vector




class BaseIterativeWriter(object):

    def __init__(self):
        self._target = None

    def set_target(self, target):
        self._target = target

    def write(self, title, contents_it):
        raise NotImplementedError()



#!/usr/bin/python
import tensorflow as tf
from arenets.context.architectures.cnn import VanillaCNN
from arenets.context.configurations.cnn import CNNConfig
from arenets.sample import InputSample


class PiecewiseCNN(VanillaCNN):
    """
    Author: Daojian Zeng, Kang Liu, Yubo Chen, Jun Zhao
    Paper: https://www.aclweb.org/anthology/D15-1203/
    Code (unofficial repo): https://github.com/nicolay-r/sentiment-pcnn
    """

    @property
    def ContextEmbeddingSize(self):
        return 3 * self.Config.FiltersCount

    def init_context_embedding(self, embedded_terms):
        embedded_terms = self.padding(embedded_terms, self.Config.WindowSize)

        bwc_line = tf.reshape(embedded_terms,
                              [self.Config.BatchSize,
                               (self.Config.TermsPerContext + (self.Config.WindowSize - 1)) * self.TermEmbeddingSize,
                               1])

        bwc_conv = tf.nn.conv1d(bwc_line, self.Hidden[self.H_conv_filter], self.TermEmbeddingSize,
                                "VALID",
                                data_format="NHWC",
                                name="conv")

        # slice all data into 3 parts -- before, inner, and after according to relation
        sliced = tf.TensorArray(dtype=tf.float32, size=self.Config.BatchSize, infer_shape=False, dynamic_size=True)
        _, _, _, _, _, sliced = tf.while_loop(
                lambda i, *_: tf.less(i, self.Config.BatchSize),
                self.__splitting,
                [0,
                 self.get_input_parameter(InputSample.I_SUBJ_IND),
                 self.get_input_parameter(InputSample.I_OBJ_IND),
                 bwc_conv,
                 self.Config.FiltersCount, sliced])

        sliced = tf.squeeze(sliced.concat())

        # Max Pooling
        bwgc_mpool = tf.nn.max_pool2d(
                input=sliced,
                ksize=[1, 1, self.Config.TermsPerContext, 1],
                strides=[1, 1, self.Config.TermsPerContext, 1],
                padding='VALID',
                data_format="NHWC")

        bwc_mpool = tf.squeeze(bwgc_mpool, [2])
        bcw_mpool = tf.transpose(bwc_mpool, perm=[0, 2, 1])
        g = tf.reshape(bcw_mpool, [self.Config.BatchSize, 3 * self.Config.FiltersCount])

        return tf.concat(g, axis=-1)

    def init_body_dependent_hidden_states(self):
        assert(isinstance(self.Config, CNNConfig))
        super(PiecewiseCNN, self).init_body_dependent_hidden_states()

        self.Hidden[self.H_W] = tf.compat.v1.get_variable(
            name='PCNN_{}'.format(self.H_W),
            shape=[self.ContextEmbeddingSize, self.Config.HiddenSize],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer,
            dtype=tf.float32)

    @staticmethod
    def __splitting(i, p_subj_ind, p_obj_ind, bwc_conv, channels_count, outputs):
        l_ind = tf.minimum(tf.gather(p_subj_ind, [i]), tf.gather(p_obj_ind, [i]))  # left
        r_ind = tf.maximum(tf.gather(p_subj_ind, [i]), tf.gather(p_obj_ind, [i]))  # right

        # total width (words count)
        # switching to get_variable requires complicated modification.
        w = tf.Variable(initial_value=bwc_conv.shape[1],
                        dtype=tf.int32)

        b_slice_from = [i, 0, 0]
        b_slice_size = tf.concat([[1], l_ind, [channels_count]], 0)
        m_slice_from = tf.concat([[i], l_ind, [0]], 0)
        m_slice_size = tf.concat([[1], r_ind - l_ind, [channels_count]], 0)
        a_slice_from = tf.concat([[i], r_ind, [0]], 0)
        a_slice_size = tf.concat([[1], w-r_ind, [channels_count]], 0)

        bwc_split_b = tf.slice(bwc_conv, b_slice_from, b_slice_size)
        bwc_split_m = tf.slice(bwc_conv, m_slice_from, m_slice_size)
        bwc_split_a = tf.slice(bwc_conv, a_slice_from, a_slice_size)

        pad_b = tf.concat([[[0, 0]],
                           tf.reshape(tf.concat([w-l_ind, [0]], 0), shape=[1, 2]),
                           [[0, 0]]],
                          axis=0)

        pad_m = tf.concat([[[0, 0]],
                           tf.reshape(tf.concat([w-r_ind+l_ind, [0]], 0), shape=[1, 2]),
                           [[0, 0]]],
                          axis=0)

        pad_a = tf.concat([[[0, 0]],
                           tf.reshape(tf.concat([r_ind, [0]], 0), shape=[1, 2]),
                           [[0, 0]]],
                          axis=0)

        bwc_split_b = tf.pad(bwc_split_b, pad_b, constant_values=tf.float32.min)
        bwc_split_m = tf.pad(bwc_split_m, pad_m, constant_values=tf.float32.min)
        bwc_split_a = tf.pad(bwc_split_a, pad_a, constant_values=tf.float32.min)

        outputs = outputs.write(i, [[bwc_split_b, bwc_split_m, bwc_split_a]])

        i += 1
        return i, p_subj_ind, p_obj_ind, bwc_conv, channels_count, outputs



from enum import Enum


class ModelNames(Enum):

    CNN = 'cnn'
    CNNAttSelfPZhou = 'cnn-self-att-pzhou'
    CNNAttSelfZYang = 'cnn-self-att-zyang'
    AttEndsCNN = 'att-cnn'
    AttEndsAndFramesCNN = 'att-ef-cnn'
    AttSynonymEndsCNN = 'att-se-cnn'
    AttSynonymEndsPCNN = 'att-se-pcnn'
    AttSynonymEndsBiLSTM = 'att-se-bilstm'
    AttSynonymEndsAndFramesCNN = 'att-sef-cnn'
    AttSynonymEndsAndFramesPCNN = 'att-sef-pcnn'
    AttSynonymEndsAndFramesBiLSTM = 'att-sef-bilstm'
    AttEndsAndFramesPCNN = 'att-ef-pcnn'
    AttEndsAndFramesBiLSTM = 'att-ef-bilstm'
    AttEndsPCNN = 'att-pcnn'
    AttFramesCNN = 'att-frames-cnn'
    AttFramesPCNN = 'att-frames-pcnn'
    SelfAttentionBiLSTM = 'self-att-bilstm'
    BiLSTM = 'bilstm'
    IANFrames = 'ian'
    IANEnds = 'ian-ends'
    IANEndsAndFrames = 'ian-ef'
    IANSynonymEnds = 'ian-se'
    IANSynonymEndsAndFrames = 'ian-sef'
    PCNN = 'pcnn'
    LSTM = 'lstm'
    RCNN = 'rcnn'
    RCNNAttPZhou = 'rcnn-att-p-zhou'
    RCNNAttZYang = 'rcnn-att-z-yang'
    AttFramesBiLSTM = 'att-frames-bilstm'
    AttSelfZYangBiLSTM = 'att-bilstm-z-yang'
    AttSelfPZhouBiLSTM = 'att-bilstm-p-zhou'


class ModelNamesService(object):

    __names = dict([(item.value, item) for item in ModelNames])

    @staticmethod
    def get_type_by_name(name):
        return ModelNamesService.__names[name]

    @staticmethod
    def iter_supported_names():
        return iter(list(ModelNamesService.__names.keys()))




from arenets.features.utils import create_filled_array


class FrameConnotationFeatures(object):

    @ staticmethod
    def to_input(frame_inds, frame_connotations, size, filler):
        assert(isinstance(frame_inds, list))
        assert(isinstance(frame_connotations, list))
        assert(len(frame_inds) == len(frame_connotations))

        vector = create_filled_array(size=size, value=filler)

        for frame_ind, frame_ind_in_sample in enumerate(frame_inds):
            if frame_ind_in_sample >= len(vector):
                continue
            vector[frame_ind_in_sample] = frame_connotations[frame_ind]

        return vector



#!/usr/bin/python
import logging
import sys
import unittest

sys.path.append('../')

from tests.pos import PartOfSpeechTypesService
from tests.tf_networks.utils import init_config
from tests.tf_networks.supported import get_supported
from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.multi.architectures.max_pooling import MaxPoolingOverSentences
from arenets.multi.configurations.base import BaseMultiInstanceConfig


class TestMultiInstanceCompile(unittest.TestCase):

    @staticmethod
    def mpmi(context_config, context_network):
        assert(isinstance(context_config, DefaultNetworkConfig))

        context_config.modify_classes_count(3)

        logging.info("TEST: {}".format(context_network))
        config = BaseMultiInstanceConfig(context_config)

        config.modify_classes_count(3)

        network = MaxPoolingOverSentences(context_network=context_network)

        pos_items_count = PartOfSpeechTypesService.get_mystem_pos_count()
        init_config(config=config,
                    pos_items_count=pos_items_count)

        network.compile(config, reset_graph=True, graph_seed=42)

    def test(self):
        logging.basicConfig(level=logging.INFO)

        for config, network in get_supported():
            self.mpmi(config, network)


if __name__ == '__main__':
    unittest.main()



from arenets.core.pipeline.item_base import EpochHandlingPipelineItem


def get_item_from_pipeline(pipeline, item_type):
    assert (isinstance(pipeline, list))
    assert (issubclass(item_type, EpochHandlingPipelineItem))

    for item in pipeline:
        if isinstance(item, item_type):
            return item



import tensorflow as tf


def self_attention_by_peng_zhou(inputs):
    """
    Attention method proposed by:
    Title: Attention-Based Bidirectional Long Short-Term Memory Networks for Relation Classification
    Authors: Peng Zhou, Wei Shi, Jun Tian, Zhenyu Qi, Bingchen Li, Hongwei Hao, Bo Xu
    Paper: https://www.aclweb.org/anthology/P16-2034
    Code author: SeoSangwoo (c), https://github.com/SeoSangwoo
    Code: https://github.com/SeoSangwoo/Attention-Based-BiLSTM-relation-extraction

    inputs: Tensor
        tensor of shape (B, T, D), where
            B -- batch, T -- time-series (context terms), D -- data or hidden embedding.
    """
    assert(isinstance(inputs, tf.Tensor))

    # Trainable parameters
    hidden_size = inputs.shape[2].value
    u_omega = tf.compat.v1.get_variable(name="u_omega",
                                        shape=[hidden_size],
                                        initializer=tf.keras.initializers.glorot_normal())

    with tf.name_scope('v'):
        v = tf.tanh(inputs)

    # For each of the timestamps its vector of size A from `v` is reduced with `u` vector
    vu = tf.tensordot(v, u_omega, axes=1, name='vu')  # (B,T) shape
    alphas = tf.nn.softmax(vu, name='alphas')  # (B,T) shape

    return alphas


def calculate_sequential_attentive_weights_by_peng_zhou(inputs, alphas):
    """ Authors: Peng Zhou, Wei Shi, Jun Tian, Zhenyu Qi, Bingchen Li, Hongwei Hao, Bo Xu
        Paper: https://www.aclweb.org/anthology/P16-2034
        Proposes to reduce the time-series parameter
        inputs: Tensor
            tensor of shape (B, T, D), where
                B -- batch, T -- time-series (context terms), D -- data or hidden embedding.
        alphas: Tensor
            tensor of shape (B, T), which corresponds to weights of the particular time-series item>
    """

    # Output of (Bi-)RNN is reduced with attention vector; the result has (B,D) shape
    output = tf.reduce_sum(inputs * tf.expand_dims(alphas, -1), 1)

    # Final output with tanh
    output = tf.tanh(output)

    return output



import logging

import numpy as np

from arenets.arekit.common.utils import create_dir_if_not_exists
from arenets.core.base_writer import BaseWriter

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class NpzDataWriter(BaseWriter):

    def write(self, data, target):
        assert(isinstance(target, str))
        logger.info("Save: {}".format(target))
        create_dir_if_not_exists(target)
        np.savez(target, data)



from os.path import join

from arenets.arekit.common.data.views.samples import LinkedSamplesStorageView
from arenets.arekit.common.data_type import DataType
from arenets.arekit.common.experiment.api.base_samples_io import BaseSamplesIO
from arenets.arekit.common.pipeline.context import PipelineContext
from arenets.arekit.common.pipeline.items.base import BasePipelineItem
from arenets.core.callback.writer import PredictResultWriterCallback
from arenets.core.ctx_inference import InferenceContext
from arenets.core.embedding_io import BaseEmbeddingIO
from arenets.core.model import BaseTensorflowModel
from arenets.core.model_ctx import TensorflowModelContext
from arenets.core.pipeline.item_fit import MinibatchFittingPipelineItem
from arenets.core.pipeline.item_keep_hidden import MinibatchHiddenFetcherPipelineItem
from arenets.core.pipeline.item_predict import EpochLabelsPredictorPipelineItem
from arenets.core.pipeline.item_predict_labeling import EpochLabelsCollectorPipelineItem
from arenets.core.predict.provider.base import BasePredictProvider
from arenets.core.writer.base_writer import BaseIterativeWriter
from arenets.factory import create_network_and_network_config_funcs
from arenets.shapes import NetworkInputShapes


class TensorflowNetworkInferencePipelineItem(BasePipelineItem):

    def __init__(self, model_name, bags_collection_type, model_input_type, predict_writer,
                 data_type, bag_size, bags_per_minibatch, nn_io, labels_count, callbacks,
                 predict_provider, modify_config_func=None, part_of_speech_types_count=100):
        assert(isinstance(callbacks, list))
        assert(isinstance(bag_size, int))
        assert(isinstance(data_type, DataType))
        assert(isinstance(predict_writer, BaseIterativeWriter))
        assert(isinstance(predict_provider, BasePredictProvider))
        assert(callable(modify_config_func) or modify_config_func is None)

        # Create network an configuration.
        network_func, config_func = create_network_and_network_config_funcs(
            model_name=model_name, model_input_type=model_input_type)

        # setup network and config parameters.
        self.__network = network_func()
        self.__config = config_func()
        self.__config.modify_classes_count(labels_count)
        self.__config.modify_bag_size(bag_size)
        self.__config.modify_bags_per_minibatch(bags_per_minibatch)
        self.__config.set_class_weights([1] * labels_count)
        self.__config.set_pos_count(part_of_speech_types_count)
        self.__config.reinit_config_dependent_parameters()

        # Custom configuration modification function.
        if modify_config_func is not None:
            modify_config_func(self.__config)

        # intialize model context.
        self.__create_model_ctx = lambda inference_ctx: TensorflowModelContext(
            nn_io=nn_io,
            network=self.__network,
            config=self.__config,
            inference_ctx=inference_ctx,
            bags_collection_type=bags_collection_type)

        self.__callbacks = callbacks + [
            PredictResultWriterCallback(labels_count=labels_count,
                                        predict_provider=predict_provider,
                                        writer=predict_writer)
        ]

        self.__writer = predict_writer
        self.__bags_collection_type = bags_collection_type
        self.__data_type = data_type

    def apply_core(self, input_data, pipeline_ctx):
        assert(isinstance(pipeline_ctx, PipelineContext))
        assert("emb_io" in input_data)
        assert("samples_io" in input_data)
        assert("predict_root" in input_data)

        emb_io = input_data["emb_io"]
        samples_io = input_data["samples_io"]
        predict_root = input_data["predict_root"]
        assert(isinstance(emb_io, BaseEmbeddingIO))
        assert(isinstance(samples_io, BaseSamplesIO))

        # Setup predicted result writer.
        full_model_name = pipeline_ctx.provide_or_none("full_model_name")
        tgt = join(predict_root, "predict-{fmn}-{dtype}.csv".format(
            fmn=full_model_name, dtype=str(self.__data_type).lower().split('.')[-1]))

        # Fetch other required in further information from input_data.
        embedding = emb_io.load_embedding()
        terms_vocab = emb_io.load_vocab()
        unknown_term_index = emb_io.UnknownTermIndex

        # Setup config parameters.
        self.__config.set_term_embedding(embedding)

        inference_ctx = InferenceContext.create_empty()
        inference_ctx.initialize(
            dtypes=[self.__data_type],
            bags_collection_type=self.__bags_collection_type,
            load_target_func=lambda data_type: samples_io.create_target(data_type=data_type),
            samples_reader=samples_io.Reader,
            samples_view=LinkedSamplesStorageView(),
            is_external_vocab=True,
            terms_vocab=terms_vocab,
            unknown_term_index=unknown_term_index,
            labels_count=self.__config.ClassesCount,
            input_shapes=NetworkInputShapes(iter_pairs=[
                (NetworkInputShapes.FRAMES_PER_CONTEXT, self.__config.FramesPerContext),
                (NetworkInputShapes.TERMS_PER_CONTEXT, self.__config.TermsPerContext),
                (NetworkInputShapes.SYNONYMS_PER_CONTEXT, self.__config.SynonymsPerContext),
            ]),
            bag_size=self.__config.BagSize)

        # Model preparation.
        model = BaseTensorflowModel(
            context=self.__create_model_ctx(inference_ctx),
            callbacks=self.__callbacks,
            predict_pipeline=[
                EpochLabelsPredictorPipelineItem(),
                EpochLabelsCollectorPipelineItem(),
                MinibatchHiddenFetcherPipelineItem()
            ],
            fit_pipeline=[MinibatchFittingPipelineItem()])

        self.__writer.set_target(tgt)

        model.predict(data_type=self.__data_type, do_compile=True)





"""
paper: Independently Recurrent Neural Network (IndRNN): Building A Longer and Deeper RNN
authors: Shuai Li, Wanqing Li, Chris Cook, Ce Zhu, Yanbo Gao
paper link: https://arxiv.org/pdf/1803.04831.pdf
code: https://github.com/batzner/indrnn
"""
# TODO. copy cell



import logging
import os
import tensorflow as tf

logger = logging.getLogger(__name__)


class TensorflowNetworkStatesProvider(object):
    """ Allows to perform tensorflow states reading and writing.
    """

    def __init__(self):
        self.__saver = None

    # region private methods

    def __init_saver(self):
        """ Note: it should be initialized once sesion has been created
            therefore we defer this action, rather than perform so at __init__
        """
        self.__saver = tf.compat.v1.train.Saver(max_to_keep=2)

    def __load_model_core(self, session, save_path):
        save_dir = os.path.dirname(save_path)

        if self.__saver is None:
            self.__init_saver()

        self.__saver.restore(sess=session, save_path=tf.train.latest_checkpoint(save_dir))

    def __save_model_core(self, session, save_path):
        if self.__saver is None:
            self.__init_saver()
        self.__saver.save(sess=session, save_path=save_path, write_meta_graph=False)

    # endregion

    def load_model(self, sess, path_tf_prefix):
        assert(isinstance(sess, tf.compat.v1.Session))

        saved_model_dir = "{}/".format(path_tf_prefix)

        if not os.path.exists(saved_model_dir):
            # Skip the case when model is not available.
            logger.info('Model was not found at: "{path}"'.format(path=saved_model_dir))
            logger.info('Skipping loading process!"')
            return

        logger.info("----------------------------------")
        logger.info("Loading Tensorflow model state: {}".format(saved_model_dir))
        logger.info("----------------------------------")
        self.__load_model_core(session=sess, save_path=saved_model_dir)

    def save_model(self, sess, path_tf_prefix):
        assert(isinstance(sess, tf.compat.v1.Session))
        logger.info("Update TensorFlow model state: {}".format(path_tf_prefix))
        self.__save_model_core(session=sess, save_path=path_tf_prefix)



import numpy as np


def in_window(window_begin, window_end, ind):
    return window_begin <= ind < window_end


def pad_right_inplace(lst, pad_size, filler):

    assert(pad_size - len(lst) > 0)
    lst.extend([filler] * (pad_size - len(lst)))


def pad_right_or_crop_inplace(lst, pad_size, filler):

    if len(lst) < pad_size:
        pad_right_inplace(lst, pad_size, filler)
    else:
        del lst[pad_size:]


def create_filled_array(size, value):
    return np.full(size, value, dtype=np.int32)


def create_zeros(size):
    return np.zeros(size, dtype=np.int32)




import logging

import numpy as np

logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.INFO)


class VocabRepositoryUtils(object):

    @staticmethod
    def load(source):
        vocab = np.loadtxt(source, dtype=str, comments=None)
        logger.info("Loading vocabulary [size={size}]: {filepath}".format(size=len(vocab), filepath=source))
        return vocab

    @staticmethod
    def save(data, target):
        logger.info("Saving vocabulary [size={size}]: {filepath}".format(size=len(data), filepath=target))
        np.savetxt(target, data, fmt='%s')


from arenets.attention.architectures.self_p_zhou import self_attention_by_peng_zhou
from arenets.context.architectures.att_self_rcnn import AttentionSelfRCNN


class AttentionSelfPZhouRCNN(AttentionSelfRCNN):

    def get_attention_alphas(self, rnn_outputs):
        return self_attention_by_peng_zhou(rnn_outputs)



from arenets.attention.architectures.self_z_yang import self_attention_by_z_yang
from arenets.context.architectures.att_self_cnn import SelfAttentionCNN


class AttentionSelfZYangCNN(SelfAttentionCNN):

    def get_attention_alphas(self, input_data):
        return self_attention_by_z_yang(inputs=input_data, attention_size=self.Config.AttentionSize)



import pandas as pd

from arenets.arekit.common.data.storages.base import BaseRowsStorage


class PandasBasedRowsStorage(BaseRowsStorage):
    """ Storage Kernel functions implementation, based on the pandas DataFrames.
    """

    def __init__(self, df=None):
        assert(isinstance(df, pd.DataFrame) or df is None)
        self._df = df

    # region protected methods

    def _iter_rows(self):
        assert(isinstance(self._df, pd.DataFrame))
        for row_index, row in self._df.iterrows():
            yield row_index, row.to_dict()

    def _get_rows_count(self):
        return len(self._df)

    # endregion



import tensorflow as tf
from arenets.context.configurations.rnn import RNNConfig
from arenets.tf_helpers.cell_types import CellTypes


class RCNNConfig(RNNConfig):

    __context_embedding_size = 300

    def __init__(self):
        super(RCNNConfig, self).__init__()
        super(RCNNConfig, self).modify_hidden_size(128)
        super(RCNNConfig, self).modify_cell_type(CellTypes.LSTM)
        super(RCNNConfig, self).modify_dropout_rnn_keep_prob(0.8)

    # region properties

    @property
    def SurroundingOneSideContextEmbeddingSize(self):
        return self.__context_embedding_size

    @property
    def WeightInitializer(self):
        return tf.contrib.layers.xavier_initializer()

    @property
    def BiasInitializer(self):
        return tf.constant_initializer(0.1)

    # endregion

    def _internal_get_parameters(self):
        parameters = super(RCNNConfig, self)._internal_get_parameters()

        parameters += [
            ("rcnn:context_size (one side)", self.SurroundingOneSideContextEmbeddingSize),
        ]

        return parameters



class BaseSamplesIO(object):
    """ Represents base experiment utils for input/output for:
        samples -- data that utilized for experiments;
        results -- evaluation of experiments.
    """

    @property
    def Reader(self):
        raise NotImplementedError()

    def create_target(self, data_type):
        """ Path for reading/viewing
        """
        raise NotImplementedError()



from arenets.context.configurations.cnn import CNNConfig


class AttentionCNNBaseConfig(CNNConfig):

    # region properties

    @property
    def AttentionModel(self):
        raise NotImplementedError()

    # endregion

    def get_attention_parameters(self):
        raise NotImplementedError()

    # region public methods

    def _internal_get_parameters(self):
        parameters = super(AttentionCNNBaseConfig, self)._internal_get_parameters()
        parameters += self.get_attention_parameters()
        return parameters

    # endregion



from arenets.arekit.common.data.input.sample import InputSampleBase
from arenets.core.feeding.bags.bag import Bag
from arenets.core.feeding.bags.collection.base import BagsCollection
from arenets.core.input.rows_parser import ParsedSampleRow


class MultiInstanceBagsCollection(BagsCollection):

    # region private methods

    @staticmethod
    def __last_bag(bags):
        return bags[-1]

    @staticmethod
    def __complete_last_bag(bags, max_bag_size, create_empty_sample_func):
        bag = MultiInstanceBagsCollection.__last_bag(bags)
        while len(bag) < max_bag_size:
            bag.add_sample(create_empty_sample_func())

    @staticmethod
    def __is_empty_last_bag(bags):
        return len(MultiInstanceBagsCollection.__last_bag(bags)) == 0

    # endregion

    @classmethod
    def _fill_bags_list_with_linked_text_opinions(cls, bags,
                                                  parsed_rows,
                                                  bag_size,
                                                  create_sample_func,
                                                  create_empty_sample_func):

        assert(isinstance(parsed_rows, list))
        assert(isinstance(bags, list))
        assert(callable(create_sample_func))

        bags.append(Bag(parsed_rows[0].UintLabel))

        for o_ind, parsed_row in enumerate(parsed_rows):
            assert(isinstance(parsed_row, ParsedSampleRow))

            if len(MultiInstanceBagsCollection.__last_bag(bags)) == bag_size:
                bags.append(Bag(uint_label=parsed_row.UintLabel))

            s = create_sample_func(parsed_row)

            # NOTE: Now we consider that the next appeared context always continues the prior.
            prior_opinion = None

            if prior_opinion is not None and not MultiInstanceBagsCollection.__is_empty_last_bag(bags):

                # NOTE: Now we consider that the next appeared context always continues the prior.
                is_continued = True

                if not is_continued:
                    MultiInstanceBagsCollection.__complete_last_bag(
                        bags=bags,
                        max_bag_size=bag_size,
                        create_empty_sample_func=create_empty_sample_func)

                    bags.append(Bag(uint_label=parsed_row.UintLabel))

            assert(isinstance(s, InputSampleBase))
            MultiInstanceBagsCollection.__last_bag(bags).add_sample(s)

        if MultiInstanceBagsCollection.__is_empty_last_bag(bags):
            del bags[-1]
            return

        MultiInstanceBagsCollection.__complete_last_bag(
            bags=bags,
            max_bag_size=bag_size,
            create_empty_sample_func=create_empty_sample_func)



import tensorflow as tf
from arenets.context.architectures.base.att_cnn_base import AttentionCNNBase
from arenets.sample import InputSample


class AttentionSynonymEndsAndFramesCNN(AttentionCNNBase):

    def get_att_input(self):
        combined = tf.concat([self.get_input_parameter(InputSample.I_SYN_SUBJ_INDS),
                              self.get_input_parameter(InputSample.I_SYN_OBJ_INDS),
                              self.get_input_parameter(InputSample.I_FRAME_INDS)],
                             axis=-1)
        return tf.contrib.framework.sort(combined, direction='DESCENDING')




import json

from arenets.arekit.common.data.storages.base import BaseRowsStorage


class JsonBasedRowsStorage(BaseRowsStorage):

    def __init__(self, rows):
        assert(isinstance(rows, list))
        self.__rows = rows

    def _iter_rows(self):
        for row_index, row in enumerate(self.__rows):
            assert(isinstance(row, str))
            yield row_index, json.loads(row)

    def _get_rows_count(self):
        return len(self.__rows)



import tensorflow as tf
from arenets.context.architectures.base.att_pcnn_base import AttentionPCNNBase
from arenets.sample import InputSample


class AttentionEndsPCNN(AttentionPCNNBase):
    """
    Attention model based on entity pair ends and
    PCNN architecture for sentence encoding.
    """

    def get_att_input(self):
        return tf.stack([self.get_input_parameter(InputSample.I_SUBJ_IND),
                         self.get_input_parameter(InputSample.I_OBJ_IND)],
                        axis=-1)



from arenets.core.feeding.batch.base import MiniBatch
from arenets.core.pipeline.item_base import EpochHandlingPipelineItem
from arenets.core.pipeline.item_predict import EpochLabelsPredictorPipelineItem
from arenets.labeling import LabeledCollection


class EpochLabelsCollectorPipelineItem(EpochHandlingPipelineItem):

    def __init__(self):
        super(EpochLabelsCollectorPipelineItem, self).__init__()
        self.__labeled_samples = None

    @property
    def LabeledSamples(self):
        return self.__labeled_samples

    def before_epoch(self, model_context, data_type):
        super(EpochLabelsCollectorPipelineItem, self).before_epoch(model_context=model_context,
                                                                   data_type=data_type)
        pairs = self._context.get_sample_id_label_pairs(data_type)
        self.__labeled_samples = LabeledCollection(uint_labeled_ids=pairs)

    def apply_core(self, input_data, pipeline_ctx):
        assert(isinstance(input_data, MiniBatch))
        assert(EpochLabelsPredictorPipelineItem.KEY in pipeline_ctx)
        uint_labels = pipeline_ctx.provide(EpochLabelsPredictorPipelineItem.KEY)

        # Apply labeling.
        for bag_index, bag in enumerate(input_data.iter_by_bags()):
            uint_label = int(uint_labels[bag_index])
            for sample in bag:
                self.__labeled_samples.assign_uint_label(uint_label, sample.ID)



from arenets.arekit.common.data.input.reader import BaseReader
from arenets.arekit.contrib.utils.data.storages.jsonl_based import JsonBasedRowsStorage


class JsonlReader(BaseReader):

    def read(self, target):
        rows = []
        with open(target, "r") as f:
            for line in f.readlines():
                rows.append(line)
        return JsonBasedRowsStorage(rows)

    def target_extension(self):
        return ".jsonl"



import collections
import logging

from arenets.arekit.common.data.input.reader import BaseReader
from arenets.arekit.common.data.storages.base import BaseRowsStorage
from arenets.arekit.common.data.views.samples import LinkedSamplesStorageView
from arenets.arekit.common.data_type import DataType
from arenets.arekit.model.labeling.stat import calculate_labels_distribution_stat
from arenets.core.feeding.bags.collection.base import BagsCollection
from arenets.core.input.rows_parser import ParsedSampleRow
from arenets.sample import InputSample

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class InferenceContext(object):

    def __init__(self, sample_label_pairs_dict, bags_collections_dict):
        assert(isinstance(sample_label_pairs_dict, dict))
        assert(isinstance(bags_collections_dict, dict))
        self.__sample_label_pairs_dict = sample_label_pairs_dict
        self.__bags_collections_dict = bags_collections_dict
        self.__train_stat_uint_labeled_sample_row_ids = None

    # region Properties

    @property
    def BagsCollections(self):
        return self.__bags_collections_dict

    @property
    def SampleIdAndLabelPairs(self):
        return self.__sample_label_pairs_dict

    @property
    def HasNormalizedWeights(self):
        return self.__train_stat_uint_labeled_sample_row_ids is not None

    # endregion

    @classmethod
    def create_empty(cls):
        return cls(sample_label_pairs_dict={}, bags_collections_dict={})

    def initialize(self, dtypes, load_target_func, samples_view, samples_reader, is_external_vocab,
                   terms_vocab, unknown_term_index, labels_count, bags_collection_type, bag_size, input_shapes):
        """
        Perform reading information from the serialized experiment inputs.
        Initializing core configuration.
        """
        assert(isinstance(dtypes, collections.Iterable))
        assert(isinstance(is_external_vocab, bool))
        assert(isinstance(labels_count, int) and labels_count > 0)
        assert(isinstance(samples_view, LinkedSamplesStorageView))
        assert(isinstance(samples_reader, BaseReader))
        assert(callable(load_target_func))

        # Reading from serialized information
        for data_type in dtypes:

            # Load Samples Storage.
            storage = samples_reader.read(load_target_func(data_type))

            # Extracting such information from serialized files.
            bags_collection = self.__read_for_data_type(
                linked_samples_iter=samples_view.iter_from_storage(storage),
                is_external_vocab=is_external_vocab,
                bags_collection_type=bags_collection_type,
                terms_vocab=terms_vocab,
                bag_size=bag_size,
                input_shapes=input_shapes,
                desc="Filling bags collection [{}]".format(data_type),
                unknown_term_index=unknown_term_index)

            uint_labeled_sample_row_ids = self.__get_labeled_sample_row_ids(storage)

            # Saving into dictionaries.
            self.__bags_collections_dict[data_type] = bags_collection
            self.__sample_label_pairs_dict[data_type] = uint_labeled_sample_row_ids

            if data_type == DataType.Train:
                self.__train_stat_uint_labeled_sample_row_ids = uint_labeled_sample_row_ids

    def calc_normalized_weights(self, labels_count):
        assert(isinstance(labels_count, int) and labels_count > 0)

        if self.__train_stat_uint_labeled_sample_row_ids is None:
            return

        normalized_label_stat, _ = calculate_labels_distribution_stat(
            uint_labeled_sample_row_ids=self.__train_stat_uint_labeled_sample_row_ids,
            classes_count=labels_count)

        return normalized_label_stat

    # region private methods

    @staticmethod
    def __read_for_data_type(linked_samples_iter, is_external_vocab, bags_collection_type,
                             terms_vocab, unknown_term_index, bag_size, input_shapes, desc=""):
        assert(issubclass(bags_collection_type, BagsCollection))

        return bags_collection_type.from_formatted_samples(
            linked_samples_iter=linked_samples_iter,
            desc=desc,
            bag_size=bag_size,
            shuffle=True,
            create_empty_sample_func=lambda: InputSample.create_empty(input_shapes),
            create_sample_func=lambda row: InputSample.create_from_parameters(
                input_sample_id=row.SampleID,
                terms=row.Terms,
                entity_inds=row.EntityInds,
                is_external_vocab=is_external_vocab,
                subj_ind=row.SubjectIndex,
                obj_ind=row.ObjectIndex,
                terms_vocab=terms_vocab,
                frame_inds=row.TextFrameVariantIndices,
                frame_connotations=row.TextFrameConnotations,
                syn_obj_inds=row.SynonymObjectInds,
                syn_subj_inds=row.SynonymSubjectInds,
                input_shapes=input_shapes,
                pos_tags=row.PartOfSpeechTags,
                unknown_term_index=unknown_term_index))

    @staticmethod
    def __get_labeled_sample_row_ids(storage):
        assert(isinstance(storage, BaseRowsStorage))
        rows_list = []
        for _, row in storage:
            labeled_row = InferenceContext.__extract_labeled_rows(row)
            rows_list.append(labeled_row)
        return rows_list

    @staticmethod
    def __extract_labeled_rows(row):
        parsed_row = ParsedSampleRow(row)
        return parsed_row.SampleID, parsed_row.UintLabel

    # endregion



ID = 'id'
LABEL = 'label'
OPINION_INDEX = 'opinion_id'

# Corresponds to fields with attitude ends. (values, STRING)
SOURCE = 'source'
TARGET = 'target'

# Corresponds to fields with attitude ends. (indices, INT)
S_IND = 's_ind'
T_IND = 't_ind'

# Entity parameters
ENTITIES = 'entities'



import collections
import logging
from os.path import exists

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def check_targets_existence(targets):
    assert (isinstance(targets, collections.Iterable))

    result = True
    for filepath in targets:
        assert(isinstance(filepath, str))

        existed = exists(filepath)
        logger.info("Check existence [{is_existed}]: {fp}".format(is_existed=existed, fp=filepath))
        if not existed:
            result = False

    return result



from arenets.context.architectures.att_ef_bilstm import AttentionEndsAndFramesBiLSTM
from arenets.context.architectures.att_ef_cnn import AttentionEndsAndFramesCNN
from arenets.context.architectures.att_ef_pcnn import AttentionEndsAndFramesPCNN
from arenets.context.architectures.att_ends_cnn import AttentionEndsCNN
from arenets.context.architectures.att_frames_bilstm import AttentionFramesBiLSTM
from arenets.context.architectures.att_frames_cnn import AttentionFramesCNN
from arenets.context.architectures.att_se_bilstm import AttentionSynonymEndsBiLSTM
from arenets.context.architectures.att_se_cnn import AttentionSynonymEndsCNN
from arenets.context.architectures.att_self_p_zhou_bilstm import AttentionSelfPZhouBiLSTM
from arenets.context.architectures.att_self_p_zhou_rcnn import AttentionSelfPZhouRCNN
from arenets.context.architectures.att_self_z_yang_bilstm import AttentionSelfZYangBiLSTM
from arenets.context.architectures.att_self_z_yang_rcnn import AttentionSelfZYangRCNN
from arenets.context.architectures.bilstm import BiLSTM
from arenets.context.architectures.ian_ef import IANEndsAndFrames
from arenets.context.architectures.ian_ends import IANEndsBased
from arenets.context.architectures.ian_frames import IANFrames
from arenets.context.architectures.ian_se import IANSynonymEndsBased
from arenets.context.architectures.att_ends_pcnn import AttentionEndsPCNN
from arenets.context.architectures.att_frames_pcnn import AttentionFramesPCNN
from arenets.context.architectures.att_se_pcnn import AttentionSynonymEndsPCNN
from arenets.context.architectures.pcnn import PiecewiseCNN
from arenets.context.architectures.rcnn import RCNN
from arenets.context.architectures.self_att_bilstm import SelfAttentionBiLSTM
from arenets.context.configurations.att_ef_bilstm import AttentionEndsAndFramesBiLSTMConfig
from arenets.context.configurations.att_ef_cnn import AttentionEndsAndFramesCNNConfig
from arenets.context.configurations.att_ef_pcnn import AttentionEndsAndFramesPCNNConfig
from arenets.context.configurations.att_self_p_zhou_bilstm import AttentionSelfPZhouBiLSTMConfig
from arenets.context.configurations.att_ends_cnn import AttentionEndsCNNConfig
from arenets.context.configurations.bilstm import BiLSTMConfig
from arenets.context.configurations.att_ends_pcnn import AttentionEndsPCNNConfig
from arenets.context.configurations.att_frames_bilstm import AttentionFramesBiLSTMConfig
from arenets.context.configurations.att_frames_cnn import AttentionFramesCNNConfig
from arenets.context.configurations.att_frames_pcnn import AttentionFramesPCNNConfig
from arenets.context.configurations.att_self_p_zhou_rcnn import AttentionSelfPZhouRCNNConfig
from arenets.context.configurations.att_self_z_yang_bilstm import AttentionSelfZYangBiLSTMConfig
from arenets.context.configurations.att_se_bilstm import AttentionSynonymEndsBiLSTMConfig
from arenets.context.configurations.att_se_cnn import AttentionSynonymEndsCNNConfig
from arenets.context.configurations.att_se_pcnn import AttentionSynonymEndsPCNNConfig
from arenets.context.configurations.att_self_z_yang_rcnn import AttentionSelfZYangRCNNConfig
from arenets.context.configurations.ian_ef import IANEndsAndFramesConfig
from arenets.context.configurations.ian_ends import IANEndsBasedConfig
from arenets.context.configurations.ian_se import IANSynonymEndsBasedConfig
from arenets.context.configurations.ian_frames import IANFramesConfig
from arenets.context.configurations.rcnn import RCNNConfig
from arenets.context.configurations.rnn import RNNConfig
from arenets.context.configurations.self_att_bilstm import SelfAttentionBiLSTMConfig
from arenets.context.configurations.cnn import CNNConfig
from arenets.context.architectures.cnn import VanillaCNN
from arenets.context.architectures.rnn import RNN


def get_supported():

    return [# Self-attention
            (SelfAttentionBiLSTMConfig(), SelfAttentionBiLSTM()),
            (AttentionSelfPZhouBiLSTMConfig(), AttentionSelfPZhouBiLSTM()),
            (AttentionSelfZYangBiLSTMConfig(), AttentionSelfZYangBiLSTM()),

            # CNN based
            (CNNConfig(), VanillaCNN()),
            (CNNConfig(), PiecewiseCNN()),

            # RNN-based
            (RNNConfig(), RNN()),
            (BiLSTMConfig(), BiLSTM()),

            # RCNN-based models (Recurrent-CNN)
            (RCNNConfig(), RCNN()),
            (AttentionSelfPZhouRCNNConfig(), AttentionSelfPZhouRCNN()),
            (AttentionSelfZYangRCNNConfig(), AttentionSelfZYangRCNN()),

            # IAN (Interactive attention networks)
            (IANFramesConfig(), IANFrames()),
            (IANEndsAndFramesConfig(), IANEndsAndFrames()),
            (IANEndsBasedConfig(), IANEndsBased()),
            (IANSynonymEndsBasedConfig(), IANSynonymEndsBased()),

            # MLP-Attention-based
            (AttentionEndsAndFramesBiLSTMConfig(), AttentionEndsAndFramesBiLSTM()),
            (AttentionFramesBiLSTMConfig(), AttentionFramesBiLSTM()),
            (AttentionSynonymEndsBiLSTMConfig(), AttentionSynonymEndsBiLSTM()),
            (AttentionEndsAndFramesPCNNConfig(), AttentionEndsAndFramesPCNN()),
            [AttentionEndsAndFramesCNNConfig(), AttentionEndsAndFramesCNN()],
            (AttentionEndsCNNConfig(), AttentionEndsCNN()),
            (AttentionEndsPCNNConfig(), AttentionEndsPCNN()),
            (AttentionSynonymEndsPCNNConfig(), AttentionSynonymEndsPCNN()),
            (AttentionSynonymEndsCNNConfig(), AttentionSynonymEndsCNN()),
            (AttentionFramesCNNConfig(), AttentionFramesCNN()),
            (AttentionFramesPCNNConfig(), AttentionFramesPCNN())]


from arenets.arekit.common.data import const
from arenets.arekit.common.data.storages.base import BaseRowsStorage


class LinkedSamplesStorageView(object):

    def iter_from_storage(self, storage):
        assert(isinstance(storage, BaseRowsStorage))
        undefined = -1

        linked = []
        current_opinion_id = undefined

        for row_index, row_dict in storage:

            # We consider that every row is unique, so we do not have linked opinions
            # in case of the absence of OPINION_INDEX parameter.
            opinion_id = row_index if const.OPINION_INDEX not in row_dict else row_dict[const.OPINION_INDEX]

            if current_opinion_id != undefined:
                if opinion_id != current_opinion_id:
                    yield linked
                    linked = []
                    current_opinion_id = opinion_id
            else:
                current_opinion_id = opinion_id

            linked.append(row_dict)

        if len(linked) > 0:
            yield linked


from arenets.arekit.common.data_type import DataType


class BaseModel(object):
    """ Base Model
    """

    def fit(self, model_params, seed):
        raise NotImplementedError()

    def predict(self, data_type=DataType.Test):
        raise NotImplementedError()



class NetworkCallback(object):

    def on_fit_started(self, operation_cancel):
        pass

    def on_predict_finished(self, pipeline):
        pass

    def on_epoch_finished(self, pipeline, operation_cancel):
        pass


from os import listdir
from os.path import join, exists


class TensorflowNeuralNetworkModelIO(object):
    """ Provides an API for saving model states
    """

    def __init__(self, model_name, target_dir=None, source_dir=None):
        """ model_name_tag: str
                general model suffix to be adopted if it is specified.
                allows to customize your model name.
            finetuned_suffix: str
                utilized to specify that the new model represents a finetuned version of
                the previously existed.
        """
        assert(isinstance(model_name, str))
        assert(isinstance(target_dir, str) or target_dir is None)
        assert(isinstance(source_dir, str) or source_dir is None)

        if isinstance(source_dir, str) and isinstance(target_dir, str):
            if source_dir == target_dir:
                raise Exception("source_dir and taget_dir could not be "
                                "exact the same value! (Not supported)")

        self.__target_dir = target_dir
        self.__model_name = model_name

        # States related parameters that allows to load an existed
        # model and provide all the related information for further
        # fine-tuning operation.
        self.__source_dir = source_dir

    @property
    def IsPretrainedStateProvided(self):
        """ NOTE: We consider that if the folder with the original model directory
            could be found and this folder is not empty in terms of the its
            contents, then we return True; False otherwise.
        """
        if self.__source_dir is not None:
            source_model_dir = self.__get_original_state_model_dir()
            if exists(source_model_dir) and len(listdir(source_model_dir)) > 0:
                return True
        return False

    # region private methods

    def __get_original_state_model_dir(self):
        return join(self.__source_dir, self.__model_name)

    # endregion

    def get_model_saving_dir(self):
        return join(self.__target_dir, self.__model_name)

    def get_model_source_path_tf_prefix(self):
        """ Path, utilized for the fine-tunning, i.e. original model state. (Loading)
        """
        return self.__get_original_state_model_dir()

    def get_model_target_path_tf_prefix(self):
        """ Provides the template for states keeping using tensorflow serializer. (Saving)
        """
        return join(self.__target_dir, self.__model_name, self.__model_name)



from os.path import join

from arenets.arekit.common.data.input.reader import BaseReader
from arenets.arekit.common.experiment.api.base_samples_io import BaseSamplesIO
from arenets.arekit.contrib.utils.io_utils.utils import check_targets_existence


class SamplesIO(BaseSamplesIO):
    """ Samples default IO utils for samples.
        Sample is a text part which include pair of attitude participants.
        This class allows to provide saver and loader for such entries, dubbed as samples.
        Samples required for machine learning training/inferring.
    """

    def __init__(self, target_dir, reader=None, prefix="sample"):
        assert(isinstance(target_dir, str))
        assert(isinstance(prefix, str))
        assert(isinstance(reader, BaseReader) or reader is None)
        self.__target_dir = target_dir
        self.__prefix = prefix
        self.__reader = reader

    # region public methods

    @property
    def Reader(self):
        return self.__reader

    def create_target(self, data_type):
        return self.__get_input_sample_target(data_type)

    def check_targets_existed(self, data_types_iter):
        for data_type in data_types_iter:

            targets = [
                self.__get_input_sample_target(data_type=data_type),
            ]

            if not check_targets_existence(targets=targets):
                return False
        return True

    # endregion

    def __get_input_sample_target(self, data_type):
        return self.__get_filepath(out_dir=self.__target_dir,
                                   template=data_type.name.lower(),
                                   prefix=self.__prefix,
                                   extension=self.__reader.target_extension())

    @staticmethod
    def __get_filepath(out_dir, template, prefix, extension):
        assert(isinstance(template, str))
        assert(isinstance(prefix, str))
        assert(isinstance(extension, str))
        return join(out_dir, "{prefix}-{template}{extension}".format(
            prefix=prefix, template=template, extension=extension))



from arenets.attention.architectures.mlp import MLPAttention
from arenets.context.architectures.base.base import SingleInstanceNeuralNetwork
from arenets.sample import InputSample


def __get_NEVS_list(ctx_network):
    """
    Helper for additional embeddings.
    Here, NEVS stands for: name, embedding, value, size list
    """
    assert(isinstance(ctx_network, SingleInstanceNeuralNetwork))

    def ctx_network_input_or_none(p_name):
        if ctx_network.has_input_parameter(p_name):
            return ctx_network.get_input_parameter(p_name)

    i_x = "I_x"
    i_pos = "I_pos"
    i_dist_obj = "I_dist_obj"
    i_dist_subj = "I_dist_subj"
    i_nearest_dist_obj = "I_nearest_dist_obj"
    i_nearest_dist_subj = "I_nearest_dist_subj"
    i_frame_connotations = "I_connotation_frames"

    return [(i_x,
             ctx_network.TermEmbedding,
             ctx_network_input_or_none(InputSample.I_X_INDS),
             ctx_network.Config.TermEmbeddingShape[1]),

            (i_pos,
             ctx_network.POSEmbedding,
             ctx_network_input_or_none(InputSample.I_POS_INDS),
             ctx_network.Config.PosEmbeddingSize),

            (i_dist_obj,
             ctx_network.DistanceEmbedding,
             ctx_network_input_or_none(InputSample.I_OBJ_DISTS),
             ctx_network.Config.DistanceEmbeddingSize),

            (i_dist_subj,
             ctx_network.DistanceEmbedding,
             ctx_network_input_or_none(InputSample.I_SUBJ_DISTS),
             ctx_network.Config.DistanceEmbeddingSize),

            (i_nearest_dist_obj,
             ctx_network.DistanceEmbedding,
             ctx_network_input_or_none(InputSample.I_NEAREST_OBJ_DISTS),
             ctx_network.Config.DistanceEmbeddingSize),

            (i_nearest_dist_subj,
             ctx_network.DistanceEmbedding,
             ctx_network_input_or_none(InputSample.I_NEAREST_SUBJ_DISTS),
             ctx_network.Config.DistanceEmbeddingSize),

            (i_frame_connotations,
             ctx_network.SentimentEmbedding,
             ctx_network_input_or_none(InputSample.I_FRAME_CONNOTATIONS),
             ctx_network.Config.SentimentEmbeddingSize)
            ]


def get_nv(ctx_network):
    return [(n, v) for n, _, v, _ in __get_NEVS_list(ctx_network)]


def get_ev(ctx_network):
    return [(e, v) for _, e, v, _ in __get_NEVS_list(ctx_network)]


def get_ne(ctx_network):
    return [(n, e) for n, e, _, _ in __get_NEVS_list(ctx_network)]


def get_ns(ctx_network):
    return [(n, s) for n, _, _, s in __get_NEVS_list(ctx_network)]


def init_mlp_attention_embedding(ctx_network, mlp_att, keys):
    assert(isinstance(ctx_network, SingleInstanceNeuralNetwork))
    assert(isinstance(mlp_att, MLPAttention))

    mlp_att.set_input(param_names_with_values=get_nv(ctx_network),
                      keys=keys)

    att_e, att_w = mlp_att.init_body(params_embeddings=get_ne(ctx_network))

    return att_e, att_w




class OperationCancellation(object):

    def __init__(self):
        self.__is_cancelled = False

    @property
    def IsCancelled(self):
        return self.__is_cancelled

    def Cancel(self):
        self.__is_cancelled = True



from arenets.arekit.model.params import BaseModelParams


class NeuralNetworkModelParams(BaseModelParams):

    def __init__(self, epochs_count):
        super(NeuralNetworkModelParams, self).__init__()
        self.__epochs_count = epochs_count

    @property
    def EpochsCount(self):
        return self.__epochs_count



import random

import numpy as np

from arenets.arekit.common.data.input.sample import InputSampleBase
from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.features.pointers import PointersFeature
from arenets.features.sample_dist import DistanceFeatures
from arenets.features.term_connotation import FrameConnotationFeatures
from arenets.features.term_indices import IndicesFeature
from arenets.features.term_types import calculate_term_types
from arenets.features.utils import pad_right_or_crop_inplace
from arenets.shapes import NetworkInputShapes


class InputSample(InputSampleBase):
    """
    Base sample which is a part of a Bag
    It provides a to_network_input method which
    generates an input info in an appropriate way
    """

    # It is important to name with 'I_' prefix
    I_X_INDS = "x_indices"
    I_SYN_SUBJ_INDS = "syn_subj_inds"
    I_SYN_OBJ_INDS = "syn_obj_inds"
    I_SUBJ_IND = "subj_inds"
    I_OBJ_IND = "obj_inds"
    I_SUBJ_DISTS = "subj_dist"
    I_NEAREST_SUBJ_DISTS = "nearest_subj_dist"
    I_NEAREST_OBJ_DISTS = "nearest_obj_dist"
    I_OBJ_DISTS = "obj_dist"
    I_POS_INDS = "pos_inds"
    I_TERM_TYPE = "term_type"
    I_FRAME_INDS = 'frame_inds'
    I_FRAME_CONNOTATIONS = 'frame_connotations'

    # TODO: Should be -1, but now it is not supported
    FRAME_SENT_ROLES_PAD_VALUE = 0
    FRAMES_PAD_VALUE = 0
    POS_PAD_VALUE = 0
    X_PAD_VALUE = 0
    TERM_TYPE_PAD_VALUE = -1
    SYNONYMS_PAD_VALUE = 0

    def __init__(self,
                 X,
                 subj_ind,
                 obj_ind,
                 syn_subj_inds,
                 syn_obj_inds,
                 dist_from_subj,
                 dist_from_obj,
                 dist_nearest_subj,
                 dist_nearest_obj,
                 pos_indices,
                 term_type,
                 frame_indices,
                 frame_connotations,
                 input_sample_id,
                 shift_index_dbg=0):
        assert(isinstance(X, np.ndarray))
        assert(isinstance(subj_ind, int))
        assert(isinstance(obj_ind, int))
        assert(isinstance(syn_obj_inds, np.ndarray))
        assert(isinstance(syn_subj_inds, np.ndarray))
        assert(isinstance(dist_from_subj, np.ndarray))
        assert(isinstance(dist_from_obj, np.ndarray))
        assert(isinstance(dist_nearest_subj, np.ndarray))
        assert(isinstance(dist_nearest_obj, np.ndarray))
        assert(isinstance(pos_indices, np.ndarray))
        assert(isinstance(term_type, np.ndarray))
        assert(isinstance(frame_indices, np.ndarray))
        assert(isinstance(frame_connotations, np.ndarray))
        assert(isinstance(shift_index_dbg, int))

        values = [(InputSample.I_X_INDS, X),
                  (InputSample.I_SUBJ_IND, subj_ind),
                  (InputSample.I_OBJ_IND, obj_ind),
                  (InputSample.I_SYN_OBJ_INDS, syn_obj_inds),
                  (InputSample.I_SYN_SUBJ_INDS, syn_subj_inds),
                  (InputSample.I_SUBJ_DISTS, dist_from_subj),
                  (InputSample.I_OBJ_DISTS, dist_from_obj),
                  (InputSample.I_NEAREST_SUBJ_DISTS, dist_nearest_subj),
                  (InputSample.I_NEAREST_OBJ_DISTS, dist_nearest_obj),
                  (InputSample.I_POS_INDS, pos_indices),
                  (InputSample.I_FRAME_INDS, frame_indices),
                  (InputSample.I_FRAME_CONNOTATIONS, frame_connotations),
                  (InputSample.I_TERM_TYPE, term_type)]

        super(InputSample, self).__init__(shift_index_dbg=shift_index_dbg,
                                          input_sample_id=input_sample_id,
                                          values=values)

    # region class methods

    @classmethod
    def create_empty(cls, input_shapes):
        assert(isinstance(input_shapes, NetworkInputShapes))

        blank_synonyms = np.zeros(input_shapes.get_shape(input_shapes.SYNONYMS_PER_CONTEXT))
        blank_terms = np.zeros(input_shapes.get_shape(input_shapes.TERMS_PER_CONTEXT))
        blank_frames = np.full(shape=input_shapes.get_shape(input_shapes.FRAMES_PER_CONTEXT),
                               fill_value=cls.FRAMES_PAD_VALUE)
        return cls(X=blank_terms,
                   subj_ind=0,
                   obj_ind=1,
                   syn_subj_inds=blank_synonyms,
                   syn_obj_inds=blank_synonyms,
                   dist_from_subj=blank_terms,
                   dist_from_obj=blank_terms,
                   pos_indices=blank_terms,
                   term_type=blank_terms,
                   dist_nearest_subj=blank_terms,
                   dist_nearest_obj=blank_terms,
                   frame_connotations=blank_terms,
                   frame_indices=blank_frames,
                   input_sample_id="1")

    # TODO. Refactoring #199.
    @classmethod
    def _generate_test(cls, config):
        assert(isinstance(config, DefaultNetworkConfig))
        blank_synonyms = np.zeros(config.SynonymsPerContext)
        blank_terms = np.random.randint(0, 3, config.TermsPerContext)
        blank_frames = np.full(shape=config.FramesPerContext,
                               fill_value=cls.FRAMES_PAD_VALUE)
        return cls(X=blank_terms,
                   subj_ind=random.randint(0, 3),
                   obj_ind=random.randint(0, 3),
                   syn_subj_inds=blank_synonyms,
                   syn_obj_inds=blank_synonyms,
                   dist_from_subj=blank_terms,
                   dist_from_obj=blank_terms,
                   pos_indices=np.random.randint(0, 5, config.TermsPerContext),
                   term_type=np.random.randint(0, 3, config.TermsPerContext),
                   dist_nearest_subj=blank_terms,
                   dist_nearest_obj=blank_terms,
                   frame_connotations=blank_terms,
                   frame_indices=blank_frames,
                   input_sample_id="1")

    @classmethod
    def __get_index_by_term(cls, term, terms_vocab, is_external_vocab, unknown_term_index):

        if not is_external_vocab:
            # Since we consider that all the existed terms presented in vocabulary
            # we obtain the related index without any additional checks
            return terms_vocab[term]

        # In case of non-native vocabulary, we consider an additional
        # placeholed when the related term has not been found in vocabulary.
        return terms_vocab[term] if term in terms_vocab else unknown_term_index

    @staticmethod
    def calc_dist_between_text_opinion_end_indices(pos1_ind, pos2_ind):
        return abs(pos1_ind - pos2_ind)

    @classmethod
    def create_from_parameters(cls,
                               input_sample_id,  # row_id
                               terms,  # list of terms, that might be found in terms_vocab
                               entity_inds,
                               is_external_vocab,
                               subj_ind,
                               obj_ind,
                               terms_vocab,  # for indexing input (all the vocabulary, obtained from offsets.py)
                               input_shapes,
                               syn_subj_inds,
                               syn_obj_inds,
                               frame_inds,
                               pos_tags,
                               frame_connotations,
                               unknown_term_index):
        """
        Here we first need to perform indexing of terms. Therefore, mark entities, frame_variants among them.
        None parameters considered as optional.
        """
        assert(isinstance(terms, list))
        assert(isinstance(terms_vocab, dict))
        assert((isinstance(subj_ind, int) and 0 <= subj_ind < len(terms)) or subj_ind is None)
        assert((isinstance(obj_ind, int) and 0 <= obj_ind < len(terms)) or obj_ind is None)
        assert(isinstance(input_shapes, NetworkInputShapes))
        assert(isinstance(entity_inds, list) or entity_inds is None)
        assert((isinstance(syn_subj_inds, list) and len(syn_subj_inds) > 0) or syn_subj_inds is None)
        assert((isinstance(syn_obj_inds, list) and len(syn_obj_inds) > 0) or syn_obj_inds is None)
        assert(isinstance(frame_inds, list) or frame_inds is None)
        assert(isinstance(pos_tags, list) or pos_tags is None)
        assert(isinstance(frame_connotations, list) or frame_connotations is None)
        assert(isinstance(unknown_term_index, int))

        def shift_index(ind):
            return ind - get_start_offset()

        def get_start_offset():
            return x_feature.StartIndex

        def get_end_offset():
            return x_feature.EndIndex

        if len(terms) < 2:
            raise Exception("AREnets does not support the input amount of tokens less than 2 due "
                            "to assignation of the [subj] and [obj] towards the different terms"
                            "in text in case of the most part of the models")

        # Setup default parameters.
        subj_ind = 0 if subj_ind is None else subj_ind
        obj_ind = 1 if obj_ind is None else obj_ind
        entity_inds = [] if entity_inds is None else entity_inds
        frame_inds = [] if frame_inds is None else frame_inds
        frame_connotations = [] if frame_connotations is None else frame_connotations
        syn_subj_inds = [subj_ind] if syn_subj_inds is None else syn_subj_inds
        syn_obj_inds = [obj_ind] if syn_obj_inds is None else syn_obj_inds
        pos_tags = [0] * len(terms) if pos_tags is None else pos_tags

        # * Check the compatibility of the provided pos_tags with respect to the given terms.
        # * Check that we do not organize the loop relation.
        assert(len(terms) == len(pos_tags))
        assert(subj_ind != obj_ind)

        # Setup entities set.
        entities_set = set(entity_inds)

        # Composing vectors
        x_indices = np.array([cls.__get_index_by_term(term, terms_vocab, is_external_vocab, unknown_term_index)
                              for term in terms])

        terms_per_context = input_shapes.get_shape(input_shapes.TERMS_PER_CONTEXT)
        synonyms_per_context = input_shapes.get_shape(input_shapes.SYNONYMS_PER_CONTEXT)
        frames_per_context = input_shapes.get_shape(input_shapes.FRAMES_PER_CONTEXT)

        # Check an ability to create sample by analyzing required window size.
        window_size = terms_per_context
        dist_between_entities = cls.calc_dist_between_text_opinion_end_indices(pos1_ind=subj_ind, pos2_ind=obj_ind)

        if not cls._check_ends_could_be_fitted_in_window(dist_between_entities, window_size):
            # In some cases we may encounter with mismatched of tpc (terms per context parameter)
            # utilized during serialization stage, and the one utilized in training process.
            # If the windows size is lower in the latter case, we need to notify in order to prevent
            # from the infinite loop.
            raise Exception("Bounds for sample_id='{sample_id}' with "
                            "positions obj={obj_ind}, subj={subj_ind} "
                            "(diff={dist}) could not be fit in window, "
                            "size of {window}".format(sample_id=input_sample_id,
                                                      obj_ind=obj_ind,
                                                      subj_ind=subj_ind,
                                                      dist=dist_between_entities,
                                                      window=window_size))

        x_feature = IndicesFeature.from_vector_to_be_fitted(
            value_vector=x_indices,
            e1_ind=subj_ind,
            e2_ind=obj_ind,
            expected_size=window_size,
            filler=cls.X_PAD_VALUE)

        pos_feature = IndicesFeature.from_vector_to_be_fitted(
            value_vector=np.array(pos_tags),
            e1_ind=subj_ind,
            e2_ind=obj_ind,
            expected_size=window_size,
            filler=cls.POS_PAD_VALUE)

        term_type_feature = IndicesFeature.from_vector_to_be_fitted(
            value_vector=calculate_term_types(terms=terms,
                                              entity_inds_set=entities_set),
            e1_ind=subj_ind,
            e2_ind=obj_ind,
            expected_size=window_size,
            filler=cls.TERM_TYPE_PAD_VALUE)

        frame_connotations_feature = IndicesFeature.from_vector_to_be_fitted(
            value_vector=FrameConnotationFeatures.to_input(frame_inds=frame_inds,
                                                           frame_connotations=frame_connotations,
                                                           size=len(terms),
                                                           filler=cls.FRAME_SENT_ROLES_PAD_VALUE),
            e1_ind=subj_ind,
            e2_ind=obj_ind,
            expected_size=window_size,
            filler=cls.FRAME_SENT_ROLES_PAD_VALUE)

        frames_feature = PointersFeature.create_shifted_and_fit(
            original_value=frame_inds,
            start_offset=get_start_offset(),
            end_offset=get_end_offset(),
            expected_size=frames_per_context,
            filler=cls.FRAMES_PAD_VALUE)

        syn_subj_inds_feature = PointersFeature.create_shifted_and_fit(
            original_value=syn_subj_inds,
            start_offset=get_start_offset(),
            end_offset=get_end_offset(),
            filler=cls.SYNONYMS_PAD_VALUE)

        syn_obj_inds_feature = PointersFeature.create_shifted_and_fit(
            original_value=syn_obj_inds,
            start_offset=get_start_offset(),
            end_offset=get_end_offset(),
            filler=cls.SYNONYMS_PAD_VALUE)

        shifted_subj_ind = shift_index(subj_ind)
        shifted_obj_ind = shift_index(obj_ind)

        dist_from_subj = DistanceFeatures.distance_feature(position=shifted_subj_ind, size=terms_per_context)
        dist_from_obj = DistanceFeatures.distance_feature(position=shifted_obj_ind, size=terms_per_context)

        dist_nearest_subj = DistanceFeatures.distance_abs_nearest_feature(
            positions=syn_subj_inds_feature.ValueVector,
            size=terms_per_context)

        dist_nearest_obj = DistanceFeatures.distance_abs_nearest_feature(
            positions=syn_obj_inds_feature.ValueVector,
            size=terms_per_context)

        pad_right_or_crop_inplace(lst=syn_subj_inds_feature.ValueVector,
                                  pad_size=synonyms_per_context,
                                  filler=cls.SYNONYMS_PAD_VALUE)

        pad_right_or_crop_inplace(lst=syn_obj_inds_feature.ValueVector,
                                  pad_size=synonyms_per_context,
                                  filler=cls.SYNONYMS_PAD_VALUE)

        return cls(X=np.array(x_feature.ValueVector),
                   subj_ind=shifted_subj_ind,
                   obj_ind=shifted_obj_ind,
                   syn_subj_inds=np.array(syn_subj_inds_feature.ValueVector),
                   syn_obj_inds=np.array(syn_obj_inds_feature.ValueVector),
                   dist_from_subj=dist_from_subj,
                   dist_from_obj=dist_from_obj,
                   dist_nearest_subj=dist_nearest_subj,
                   dist_nearest_obj=dist_nearest_obj,
                   pos_indices=np.array(pos_feature.ValueVector),
                   term_type=np.array(term_type_feature.ValueVector),
                   frame_indices=np.array(frames_feature.ValueVector),
                   frame_connotations=np.array(frame_connotations_feature.ValueVector),
                   input_sample_id=input_sample_id,
                   shift_index_dbg=get_start_offset())

    # endregion

    @staticmethod
    def iter_parameters():
        for var_name in dir(InputSample):
            if not var_name.startswith('I_'):
                continue
            yield getattr(InputSample, var_name)



import tensorflow as tf
from arenets.context.configurations.bilstm import BiLSTMConfig


class SelfAttentionBiLSTMConfig(BiLSTMConfig):
    """
    Paper: https://arxiv.org/abs/1703.03130
    Code Author: roomylee, https://github.com/roomylee (C)
    Code: https://github.com/roomylee/self-attentive-emb-tf
    """

    def __init__(self):
        """
        d_a_size: int
            Size of W_s1 embedding
        r_size: int
            Size of W_s2 embedding
        fc_size: int
            Size of fully connected laye
        p_coef: int
            Coefficient for penalty
        """
        super(SelfAttentionBiLSTMConfig, self).__init__()
        super(SelfAttentionBiLSTMConfig, self).modify_weight_initializer(tf.contrib.layers.xavier_initializer())
        super(SelfAttentionBiLSTMConfig, self).modify_bias_initializer(tf.constant_initializer(0.1))
        super(SelfAttentionBiLSTMConfig, self).modify_learning_rate(0.1)

        self.__fc_size = 200
        self.__r_size = 30
        self.__d_a_size = 350
        self.__p_coef = 1.0

    # region public methods

    @property
    def FullyConnectionSize(self):
        return self.__fc_size

    @property
    def RSize(self):
        return self.__r_size

    @property
    def PenaltizationTermCoef(self):
        return self.__p_coef

    @property
    def DASize(self):
        return self.__d_a_size

    # endregion

    def modify_penaltization_term_coef(self, value):
        self.__p_coef = value

    def _internal_get_parameters(self):
        parameters = super(SelfAttentionBiLSTMConfig, self)._internal_get_parameters()

        parameters += [
            ("sa-bilstm:fully_connection_size", self.FullyConnectionSize),
            ("sa-bilstm:penalty_term_coef", self.PenaltizationTermCoef),
            ("sa-bilstm:da_size (size of ws_1)", self.DASize),
            ("sa-bilstm:r_size (size of ws_2)", self.RSize)
        ]

        return parameters



import gc
import logging
import os

from arenets.arekit.common.data.views.samples import LinkedSamplesStorageView
from arenets.arekit.common.data_type import DataType
from arenets.arekit.common.pipeline.context import PipelineContext
from arenets.arekit.common.pipeline.items.base import BasePipelineItem
from arenets.arekit.contrib.utils.io_utils.embedding import NpEmbeddingIO
from arenets.arekit.contrib.utils.io_utils.samples import SamplesIO
from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.core.ctx_inference import InferenceContext
from arenets.core.feeding.bags.collection.base import BagsCollection
from arenets.core.model import BaseTensorflowModel
from arenets.core.model_ctx import TensorflowModelContext
from arenets.core.params import NeuralNetworkModelParams
from arenets.core.pipeline.item_fit import MinibatchFittingPipelineItem
from arenets.core.pipeline.item_keep_hidden import MinibatchHiddenFetcherPipelineItem
from arenets.core.pipeline.item_predict import EpochLabelsPredictorPipelineItem
from arenets.core.pipeline.item_predict_labeling import EpochLabelsCollectorPipelineItem
from arenets.shapes import NetworkInputShapes
from arenets.utils import rm_dir_contents


class NetworksTrainingPipelineItem(BasePipelineItem):

    def __init__(self, bags_collection_type, model_io, samples_io, emb_io,
                 config, create_network_func, training_epochs,
                 labels_count, network_callbacks, prepare_model_root=True, seed=None):
        assert(callable(create_network_func))
        assert(isinstance(samples_io, SamplesIO))
        assert(isinstance(emb_io, NpEmbeddingIO))
        assert(isinstance(config, DefaultNetworkConfig))
        assert(issubclass(bags_collection_type, BagsCollection))
        assert(isinstance(seed, int) or seed is None)
        assert(isinstance(training_epochs, int))
        assert(isinstance(network_callbacks, list))
        assert(isinstance(labels_count, int))

        super(NetworksTrainingPipelineItem, self).__init__()

        self.__logger = self.__create_logger()
        self.__samples_io = samples_io
        self.__emb_io = emb_io
        self.__clear_model_root_before_experiment = prepare_model_root
        self.__config = config
        self.__create_network_func = create_network_func
        self.__bags_collection_type = bags_collection_type
        self.__network_callbacks = network_callbacks
        self.__training_epochs = training_epochs
        self.__labels_count = labels_count
        self.__model_io = model_io
        self.__seed = seed

    @staticmethod
    def __create_logger():
        stream_handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s %(levelname)8s %(name)s | %(message)s')
        stream_handler.setFormatter(formatter)
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
        logger.addHandler(stream_handler)
        return logger

    def __prepare_model(self):
        # Clear model root before training optionally
        if self.__clear_model_root_before_experiment:
            rm_dir_contents(dir_path=self.__model_io.get_model_saving_dir(),
                            logger=self.__logger)

        # Disable tensorflow logging
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

        # Notify other subscribers that initialization process has been completed.
        self.__config.init_initializers()

    def __check_targets_existed(self, data_types_iter):
        """ Check that all the required resources existed.
        """

        if not self.__samples_io.check_targets_existed(data_types_iter=data_types_iter):
            return False

        if not self.__emb_io.check_targets_existed():
            return False

        return True

    def __handle_iteration(self, supported_data_types, data_type):
        assert(isinstance(supported_data_types, list))
        assert(isinstance(data_type, DataType))

        targets_existed = self.__check_targets_existed(data_types_iter=iter(supported_data_types))

        if not targets_existed:
            raise Exception("Data has not been initialized/serialized!")

        # Reading embedding.
        embedding_data = self.__emb_io.load_embedding()
        self.__config.set_term_embedding(embedding_data)
        self.__samples_io.create_target(data_type=data_type)

        # Performing samples reading process.
        inference_ctx = InferenceContext.create_empty()
        inference_ctx.initialize(
            dtypes=iter(supported_data_types),
            load_target_func=lambda dtype: self.__samples_io.create_target(data_type=dtype),
            samples_view=LinkedSamplesStorageView(),
            samples_reader=self.__samples_io.Reader,
            is_external_vocab=not self.__model_io.IsPretrainedStateProvided,
            labels_count=self.__labels_count,
            terms_vocab=self.__emb_io.load_vocab(),
            unknown_term_index=self.__emb_io.UnknownTermIndex,
            bags_collection_type=self.__bags_collection_type,
            input_shapes=NetworkInputShapes(iter_pairs=[
                (NetworkInputShapes.FRAMES_PER_CONTEXT, self.__config.FramesPerContext),
                (NetworkInputShapes.TERMS_PER_CONTEXT, self.__config.TermsPerContext),
                (NetworkInputShapes.SYNONYMS_PER_CONTEXT, self.__config.SynonymsPerContext),
            ]),
            bag_size=self.__config.BagSize)

        if inference_ctx.HasNormalizedWeights:
            weights = inference_ctx.calc_normalized_weights(labels_count=self.__labels_count)
            self.__config.set_class_weights(weights)

        # Update parameters after iteration preparation has been completed.
        self.__config.reinit_config_dependent_parameters()

        # Initialize network and model.
        network = self.__create_network_func()
        model = BaseTensorflowModel(
            context=TensorflowModelContext(
                network=network,
                config=self.__config,
                inference_ctx=inference_ctx,
                bags_collection_type=self.__bags_collection_type,
                nn_io=self.__model_io),
            callbacks=self.__network_callbacks,
            predict_pipeline=[
                EpochLabelsPredictorPipelineItem(),
                EpochLabelsCollectorPipelineItem(),
                MinibatchHiddenFetcherPipelineItem()
            ],
            fit_pipeline=[
                MinibatchFittingPipelineItem(),
                MinibatchHiddenFetcherPipelineItem()
            ])

        # Initialize model params instance.
        model_params = NeuralNetworkModelParams(epochs_count=self.__training_epochs)

        model.fit(model_params=model_params, seed=self.__seed, data_type=data_type)

        del network
        del model

        gc.collect()

    def apply_core(self, input_data, pipeline_ctx):
        assert(isinstance(pipeline_ctx, PipelineContext))
        assert("supported_data_types" in pipeline_ctx)

        # Prepare all the required data.
        self.__prepare_model()
        supported_data_types = pipeline_ctx.provide("supported_data_types")
        data_type = pipeline_ctx.provide_or_none("data_type")

        self.__handle_iteration(supported_data_types=supported_data_types,
                                data_type=data_type if data_type is not None else DataType.Train)



from enum import Enum


class DataType(Enum):
    """
    Describes collection types that supports in
    current implementation, and provides by collections.
    """

    Train = 1

    Test = 2

    Dev = 3

    Etalon = 4



from enum import Enum


class ModelInputType(Enum):
    SingleInstance = 'ctx'
    MultiInstanceMaxPooling = 'mi-mp'
    MultiInstanceWithSelfAttention = 'mi-self-att'


class ModelInputTypeService(object):

    __names = dict([(item.value, item) for item in ModelInputType])

    @staticmethod
    def __iter_supported_names():
        return iter(list(ModelInputTypeService.__names.keys()))

    @staticmethod
    def get_type_by_name(name):
        return ModelInputTypeService.__names[name]

    @staticmethod
    def find_name_by_type(input_type):
        assert(isinstance(input_type, ModelInputType))

        for name in ModelInputTypeService.__iter_supported_names():
            related_type = ModelInputTypeService.__names[name]
            if related_type == input_type:
                return name

    @staticmethod
    def iter_supported_names():
        return ModelInputTypeService.__iter_supported_names()



from setuptools import (
    setup,
    find_packages,
)


def get_requirements(filenames):
    r_total = []
    for filename in filenames:
        with open(filename) as f:
            r_local = f.read().splitlines()
            r_total.extend(r_local)
    return r_total


setup(
    name='arenets',
    version='0.23.1',
    python_requires=">=3.6, <3.7",
    description='Tensorflow-based framework which lists implementation of conventional neural '
                'network models (CNN, RNN-based) for Relation Extraction classification tasks '
                'as well as API for custom model implementation',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/nicolay-r/AREnets',
    author='Nicolay Rusnachenko',
    author_email='rusnicolay@gmail.com',
    license='MIT License',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
    keywords='natural language processing, relation extraction, sentiment analysis',
    packages=find_packages(),
    install_requires=get_requirements([
        'dependencies.txt'
    ]),
    data_files=["logo.png"],
)


import datetime

import numpy as np
import tensorflow as tf


class DefaultNetworkConfig(object):

    # region private settings

    __use_class_weights = True

    __dropout_keep_prob = 0.5
    __embedding_dropout_keep_prob = 1.0

    __classes_count = None
    __pos_count = None
    __terms_per_context = 50
    __synonyms_per_context = 3
    __frames_per_context = 5
    __bags_per_minibatch = 6
    __bag_size = 1
    __learning_rate = 0.1

    # Assumes to be initialized after all settings will be declared.
    __default_weight_initializer = None
    __default_bias_initializer = None
    __default_regularizer = None
    __default_embedding_initializer = None
    __optimiser = None

    __term_embedding_matrix = None   # Includes embeddings of: words, entities, tokens.
    __class_weights = None
    __use_pos_emb = True
    __pos_emb_size = 5
    __sent_emb_size = 5
    __dist_emb_size = 5

    __use_entity_types_in_embedding = True              # Affects on result embedding of related entity: entity + type.
    __use_entity_types_as_context_feature = False       # TODO. Context based feature, i.e. declared for all terms

    __l2_reg = 0.0

    # endregion

    def __init__(self):
        pass

    # region properties

    @property
    def L2Reg(self):
        return self.__l2_reg

    @property
    def DistanceEmbeddingSize(self):
        return self.__dist_emb_size

    @property
    def TermEmbeddingMatrix(self):
        return self.__term_embedding_matrix

    @property
    def TermEmbeddingShape(self):
        return self.__term_embedding_matrix.shape

    @property
    def TotalAmountOfTermsInEmbedding(self):
        """
        Returns vocabulary size -- total amount of words/parsed_news,
        for which embedding has been provided
        """
        return self.TermEmbeddingShape(0)

    @property
    def BiasInitializer(self):
        return self.__default_bias_initializer

    @property
    def WeightInitializer(self):
        return self.__default_weight_initializer

    @property
    def LayerRegularizer(self):
        return self.__default_regularizer

    @property
    def EmbeddingInitializer(self):
        return self.__default_embedding_initializer

    # endregion

    # region public methods

    def modify_use_entity_types_in_embedding(self, value):
        assert(isinstance(value, bool))
        self.__use_entity_types_in_embedding = value

    def modify_use_entity_types_as_context_feature(self, value):
        assert(isinstance(value, bool))
        self.__use_entity_types_as_context_feature = value

    def modify_terms_per_context(self, value):
        assert(isinstance(value, int) and value > 0)
        self.__terms_per_context = value

    def modify_synonyms_per_context(self, value):
        assert(isinstance(value, int) and value > 0)
        self.__synonyms_per_context = value

    def modify_l2_reg(self, value):
        assert(isinstance(value, float))
        self.__l2_reg = value

    def modify_weight_initializer(self, value):
        self.__default_weight_initializer = value

    def modify_bias_initializer(self, value):
        self.__default_bias_initializer = value

    def modify_regularizer(self, value):
        self.__default_regularizer = value

    def modify_classes_count(self, value):
        assert(isinstance(value, int))
        self.__classes_count = value
        self.__class_weights = None

    def modify_learning_rate(self, value):
        assert(isinstance(value, float))
        self.__learning_rate = value

    def modify_use_class_weights(self, value):
        assert(isinstance(value, bool))
        self.__use_class_weights = value

    def modify_dropout_keep_prob(self, value):
        assert(isinstance(value, float))
        assert(0 < value <= 1.0)
        self.__dropout_keep_prob = value

    def modify_embedding_dropout_keep_prob(self, value):
        assert(isinstance(value, float))
        assert(0 < value <= 1.0)
        self.__embedding_dropout_keep_prob = value

    def modify_bags_per_minibatch(self, value):
        assert(isinstance(value, int))
        self.__bags_per_minibatch = value

    def modify_bag_size(self, value):
        assert(isinstance(value, int))
        self.__bag_size = value

    def set_term_embedding(self, embedding_matrix):
        assert(isinstance(embedding_matrix, np.ndarray))
        self.__term_embedding_matrix = embedding_matrix

    def set_class_weights(self, class_weights):
        assert(isinstance(class_weights, list))
        assert(len(class_weights) == self.__classes_count)
        self.__class_weights = class_weights

    def set_pos_count(self, value):
        self.__pos_count = value

    def reinit_config_dependent_parameters(self):
        pass

    def init_initializers(self):
        assert(self.__optimiser is None)
        assert(self.__default_regularizer is None)
        assert(self.__default_embedding_initializer is None)

        # Initialize optimizer.
        self.__optimiser = self._create_optimizer()

        # Initialize default l2-regularizer.
        self.__default_regularizer = tf.contrib.layers.l2_regularizer(self.L2Reg)

        # Initialize default embedding initializer.
        self.__default_embedding_initializer = tf.contrib.layers.xavier_initializer()

        # Initializing default (optionally).
        if self.__default_weight_initializer is None:
            self.__default_weight_initializer = tf.contrib.layers.xavier_initializer()

        # Initialize bias initializer (optionally).
        if self.__default_bias_initializer is None:
            self.__default_bias_initializer = tf.zeros_initializer(dtype=tf.float32)

    # endregion

    # region properties

    @property
    def ClassesCount(self):
        return self.__classes_count

    @property
    def PosCount(self):
        return self.__pos_count

    @property
    def ClassWeights(self):
        return self.__class_weights

    @property
    def Optimiser(self):
        return self.__optimiser

    @property
    def BatchSize(self):
        return self.BagSize * self.BagsPerMinibatch

    @property
    def BagSize(self):
        return self.__bag_size

    @property
    def BagsPerMinibatch(self):
        return self.__bags_per_minibatch

    @property
    def DropoutKeepProb(self):
        return self.__dropout_keep_prob

    @property
    def TermsPerContext(self):
        return self.__terms_per_context

    @property
    def SynonymsPerContext(self):
        return self.__synonyms_per_context

    @property
    def UseClassWeights(self):
        return self.__use_class_weights

    @property
    def UsePOSEmbedding(self):
        return self.__use_pos_emb

    @property
    def PosEmbeddingSize(self):
        return self.__pos_emb_size

    @property
    def SentimentEmbeddingSize(self):
        return self.__sent_emb_size

    @property
    def LearningRate(self):
        return self.__learning_rate

    @property
    def EmbeddingDropoutKeepProb(self):
        return self.__embedding_dropout_keep_prob

    @property
    def FramesPerContext(self):
        return self.__frames_per_context

    @property
    def UseEntityTypeAsContextFeature(self):
        return self.__use_entity_types_as_context_feature

    # endregion

    def _create_optimizer(self):
        return tf.compat.v1.train.AdadeltaOptimizer(learning_rate=self.__learning_rate,
                                                    epsilon=10e-6,
                                                    rho=0.95)

    def _internal_get_parameters(self):
        return [
            ("base:current_time", datetime.datetime.now()),
            ("base:use_class_weights", self.UseClassWeights),
            ("base:dropout (keep prob)", self.DropoutKeepProb),
            ("base:classes_count", self.ClassesCount),
            ("base:class_weights", self.ClassWeights),
            ("base:terms_per_context", self.TermsPerContext),
            ("base:synonyms_per_context", self.SynonymsPerContext),
            ("base:bags_per_minibatch", self.BagsPerMinibatch),
            ("base:bag_size", self.BagSize),
            ("base:batch_size", self.BatchSize),
            ("base:use_pos_emb", self.UsePOSEmbedding),
            ("base:pos_emb_size", self.PosEmbeddingSize),
            ("base:sentiment_emb_size", self.SentimentEmbeddingSize),
            ("base:dist_embedding_size", self.DistanceEmbeddingSize),
            ("base:use_entity_types_as_context_feature", self.UseEntityTypeAsContextFeature),
            ("base:embedding dropout (keep prob)", self.EmbeddingDropoutKeepProb),
            ("base:optimizer", self.Optimiser),
            ("base:learning_rate", self.LearningRate),
            ("base:l2_reg", self.L2Reg),
            ("base:layer_regularizer", self.LayerRegularizer),
            ("base:weight_initializer", self.WeightInitializer),
            ("base:bias_initializer", self.BiasInitializer),
        ]

    def get_parameters(self):
        return [list(p) for p in self._internal_get_parameters()]



from arenets.attention.architectures.mlp_interactive import InteractiveMLPAttention
from arenets.attention.configurations.mlp_interactive import InteractiveMLPAttentionConfig
from arenets.context.configurations.base.att_cnn_base import AttentionCNNBaseConfig


class AttentionEndsAndFramesCNNConfig(AttentionCNNBaseConfig):

    def __init__(self):
        super(AttentionEndsAndFramesCNNConfig, self).__init__()
        self.__attention = None
        self.__attention_config = InteractiveMLPAttentionConfig(self.FramesPerContext + 2)

    # region properties

    @property
    def AttentionModel(self):
        return self.__attention

    # endregion

    # region public methods

    def get_attention_parameters(self):
        return self.__attention_config.get_parameters()

    def reinit_config_dependent_parameters(self):
        super(AttentionEndsAndFramesCNNConfig, self).reinit_config_dependent_parameters()

        self.__attention = InteractiveMLPAttention(
            cfg=self.__attention_config,
            batch_size=self.BatchSize,
            terms_per_context=self.TermsPerContext,
            support_zero_length=False)

    # endregion



import logging
import tensorflow as tf

from arenets.attention.architectures.mlp import MLPAttention
from arenets.sample import InputSample
from arenets.tf_helpers import filtering, sequence

logger = logging.getLogger(__name__)


class InteractiveMLPAttention(MLPAttention):
    """
    Averaging result attention of MultiLayerPerceptronAttention
    """

    def __init__(self, cfg,
                 batch_size,
                 terms_per_context,
                 support_zero_length):
        super(InteractiveMLPAttention, self).__init__(cfg, batch_size, terms_per_context)
        assert(isinstance(support_zero_length, bool))

        self.__dynamic_lens = None
        self.__support_zero_length = support_zero_length

    @property
    def AttentionEmbeddingSize(self):
        """
        Considered single entity as an average of the result
        """
        return self.TermEmbeddingSize

    @property
    def Handler(self):
        if self.__support_zero_length:
            return InteractiveMLPAttention.crop_elements_by_lengths_and_reduce_mean_with_zero_level_support
        else:
            return InteractiveMLPAttention.crop_elements_by_lengths_and_reduce_mean

    def calculate_keys_length_func(self, keys):
        scalar_length = super(InteractiveMLPAttention, self).calculate_keys_length_func(keys)

        self.__dynamic_lens = sequence.calculate_sequence_length(
            sequence=keys,
            is_neg_placeholder=InputSample.FRAMES_PAD_VALUE < 0)

        return scalar_length

    def reshape_att_sum(self, att_sum):
        """
        att_sum: [batch_size, entity_per_context, term_embedding_size]
        """
        _att_sum = tf.reshape(att_sum, shape=[self.BatchSize,
                                              self.Config.KeysPerContext,
                                              self.TermEmbeddingSize])

        mean_sum = filtering.filter_batch_elements(
            elements_type=tf.float32,
            elements=_att_sum,
            inds=self.__dynamic_lens,
            handler=self.Handler)

        return super(InteractiveMLPAttention, self).reshape_att_sum(mean_sum)

    def reshape_att_weights(self, att_weights):
        """
        att_sum: [batch_size, entity_per_context, terms_per_context]
        """
        _att_weights = tf.reshape(att_weights, shape=[self.BatchSize,
                                                      self.Config.KeysPerContext,
                                                      self.TermsPerContext])

        mean_sum = filtering.filter_batch_elements(
            elements_type=tf.float32,
            elements=_att_weights,
            inds=self.__dynamic_lens,
            handler=self.Handler)

        return tf.reshape(mean_sum, shape=[self.BatchSize, self.TermsPerContext])

    @staticmethod
    def __core_process_batch_element(i, elements, lens):
        """
        Crop all elements by calculated length and perform reduce_mean operation.

        elements: [batch, entity_per_context, term_embedding_size]
        lens: [batch, 1]
        extra_parameter: bool
            used for flag support_zero_length
        """
        row_elements = tf.squeeze(tf.gather(elements, [i], axis=0))  # [entity_per_context, term_embedding_size]
        term_embedding_size = row_elements.shape[-1]
        row_len_original = tf.reshape(tf.gather(lens, [i], axis=0), [])  # scalar
        row_frames = tf.slice(row_elements, begin=[0, 0], size=[row_len_original, term_embedding_size])

        return row_frames, term_embedding_size, row_len_original

    @staticmethod
    def crop_elements_by_lengths_and_reduce_mean_with_zero_level_support(i, elements, lens, filtered):
        """
        This handler supports the case of empty list of elements by providing a normal_dist vector in case of
        an empty amount of elements.

        It could be utilized for frames, as amount of frames per context may vary.
        The latter might not be presented in context.
        """
        logger.info("Compile model with logger which supports lens == 0")

        row_frames, term_embedding_size, row_len_original = InteractiveMLPAttention.__core_process_batch_element(
            i=i,
            elements=elements,
            lens=lens)

        zeros = tf.zeros(shape=[1, term_embedding_size])
        non_empty_frames = tf.concat(values=[row_frames, zeros], axis=0)

        # This is a correction as we have a following equation:
        # l -- denotes a row_length_original
        #    x      y         l+1
        #    -  == --- => y = --- x
        #    l     l+1         l

        scalar = tf.divide(tf.add(row_len_original, 1), tf.maximum(row_len_original, 1))
        scalar = tf.cast(scalar, dtype=tf.float32)
        row_frames = tf.scalar_mul(scalar=scalar, x=non_empty_frames)

        mean = tf.reduce_mean(row_frames, axis=0)  # result: [terms_embedding_size]
        result = tf.reshape(mean, shape=[term_embedding_size])

        return (i + 1,
                elements,
                lens,
                filtered.write(i, tf.squeeze(result)))

    @staticmethod
    def crop_elements_by_lengths_and_reduce_mean(i, elements, lens, filtered):
        """
        This is a general handler.

        Handler with the following limitation:
        - lens values > 0
        """
        logger.info("Compile model with logger which has a limitation (lens > 0)")

        row_frames, term_embedding_size, _ = InteractiveMLPAttention.__core_process_batch_element(
            i=i, elements=elements, lens=lens)

        mean = tf.reduce_mean(row_frames, axis=0)  # result: [terms_embedding_size]
        result = tf.reshape(mean, shape=[term_embedding_size])

        return (i + 1,
                elements,
                lens,
                filtered.write(i, tf.squeeze(result)))



import tensorflow as tf
from arenets.context.configurations.rnn import RNNConfig
from arenets.tf_helpers.cell_types import CellTypes


class StatesAggregationModes:

    AVERAGE = 'avg'
    LAST_IN_SEQUENCE = 'last'


class IANBaseConfig(RNNConfig):

    __states_aggregation_mode = None

    def __init__(self):
        super(IANBaseConfig, self).__init__()

        # Reinitialize default parameters.
        super(IANBaseConfig, self).modify_bias_initializer(tf.zeros_initializer())
        super(IANBaseConfig, self).modify_weight_initializer(tf.random_uniform_initializer(-0.1, 0.1))
        super(IANBaseConfig, self).modify_l2_reg(0.001)
        super(IANBaseConfig, self).modify_cell_type(CellTypes.LSTM)
        super(IANBaseConfig, self).modify_hidden_size(128)
        super(IANBaseConfig, self).modify_embedding_dropout_keep_prob(0.8)

        self.__states_aggregation_mode = StatesAggregationModes.AVERAGE

    # region Properties

    @property
    def MaxAspectLength(self):
        return self.FramesPerContext

    @property
    def MaxContextLength(self):
        return self.TermsPerContext

    @property
    def StatesAggregationMode(self):
        return self.__states_aggregation_mode

    # endregion

    # region public methods

    def modify_states_aggregation_mode(self, value):
        assert(isinstance(value, str))
        self.__states_aggregation_mode = value

    def _internal_get_parameters(self):
        parameters = super(IANBaseConfig, self)._internal_get_parameters()

        parameters += [
            ("ian:max_aspect_len", self.MaxAspectLength),
            ("ian:max_context_len", self.MaxContextLength),
            ("ian:aggregation_mode", self.StatesAggregationMode)
        ]

        return parameters

    # endregion



import tensorflow as tf
from arenets.attention import common
from arenets.context.architectures.cnn import VanillaCNN


class SelfAttentionCNN(VanillaCNN):
    """ Self-attention application, based on the
        convolved input information by a set of filters
    """

    def __init__(self):
        super(SelfAttentionCNN, self).__init__()
        self.__att_alphas = None

    def get_attention_alphas(self, input_data):
        raise NotImplementedError()

    # region public methods

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(SelfAttentionCNN, self).iter_input_dependent_hidden_parameters():
            yield name, value

        yield common.ATTENTION_WEIGHTS_LOG_PARAMETER, self.__att_alphas

    # endregion

    def convolved_transformation_optional(self, value):
        self.__att_alphas = self.get_attention_alphas(value)
        return value * tf.expand_dims(self.__att_alphas, -1)



import math
from collections import OrderedDict

import numpy as np


class NetworkInputDependentVariables:

    __name_separator = '>'

    def __init__(self):
        self.__text_opinion_ids = OrderedDict()
        self.__by_param_names = OrderedDict()

    # region public methods

    def add_input_dependent_values(self, names_list, tensor_values_list, text_opinion_ids,
                                   bags_per_minibatch,
                                   bag_size):
        """
        names: list
            list of string names of related 'tensor_values'
        tensor_values: list
            list of values with shape:
              * Sample based: [len(names), bags_per_minibatch, bag_size, vector_size]
              * Bags based (labels): [len(names), bags_per_minibatch, 1, 1]
        text_opinion_ids: list
            list of ids
        """
        assert(isinstance(names_list, list) and len(names_list) > 0)
        assert(isinstance(tensor_values_list, list) and len(tensor_values_list) > 0)
        assert(isinstance(text_opinion_ids, list))
        assert(isinstance(bags_per_minibatch, int))
        assert(isinstance(bag_size, int))
        assert(len(tensor_values_list) == len(names_list))

        for name_ind, name in enumerate(names_list):
            assert(isinstance(name, str))

            if name not in self.__by_param_names:
                self.__by_param_names[name] = []
                self.__text_opinion_ids[name] = []

            if bags_per_minibatch * bag_size != len(text_opinion_ids):
                raise Exception("values_list of '{}' has size {} != {} (text_opinion_inds length)".format(
                    name, bags_per_minibatch * bag_size, len(text_opinion_ids)))

            values_list = np.array(tensor_values_list[name_ind])
            values_list = values_list.flatten()
            if len(values_list) > bags_per_minibatch:
                mbatches_count = math.trunc(len(values_list) / (bags_per_minibatch * bag_size))
                values_list = values_list.reshape([bags_per_minibatch, bag_size, mbatches_count])
            else:
                # labels.
                values_list = values_list.reshape([bags_per_minibatch, 1, 1])

            # Save only first sentence ref.
            t_ind = 0
            for b_ind in range(bags_per_minibatch):

                v = values_list[b_ind]

                value = []
                names = []
                for s_index in range(len(v)):
                    value.append(v[s_index])
                    names.append(str(text_opinion_ids[t_ind]))
                    t_ind += 1

                self.__by_param_names[name].append(value)
                self.__text_opinion_ids[name].append(self.__name_separator.join(names))

    def iter_by_parameter_values(self, param_name):
        for i, tensor_value in enumerate(self.__by_param_names[param_name]):
            yield self.__text_opinion_ids[param_name][i], tensor_value

    def iter_var_names(self):
        return iter(self.__by_param_names.keys())

    # endregion


import tensorflow as tf
from arenets.attention import common
from arenets.context.architectures.rcnn import RCNN


class AttentionSelfRCNN(RCNN):

    def __init__(self):
        super(AttentionSelfRCNN, self).__init__()
        self.__att_alphas = None

    def get_attention_alphas(self, rnn_outputs):
        raise NotImplementedError()

    # region public methods

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(AttentionSelfRCNN, self).iter_input_dependent_hidden_parameters():
            yield name, value

        yield common.ATTENTION_WEIGHTS_LOG_PARAMETER, self.__att_alphas

    # endregion

    def modify_rnn_outputs_optional(self, output_fw, output_bw):
        rnn_outputs = tf.add(output_fw, output_bw)

        self.__att_alphas = self.get_attention_alphas(rnn_outputs)

        output_fw_w = output_fw * tf.expand_dims(self.__att_alphas, -1)
        output_bw_w = output_bw * tf.expand_dims(self.__att_alphas, -1)

        return output_fw_w, output_bw_w



import collections


class NetworkInputShapes(object):

    SYNONYMS_PER_CONTEXT = "synonyms_per_context"
    TERMS_PER_CONTEXT = "terms_per_context"
    FRAMES_PER_CONTEXT = "frames_per_context"

    def __init__(self, iter_pairs):
        assert(isinstance(iter_pairs, collections.Iterable))
        self.__d = dict()
        for key, value in iter_pairs:
            self.__d[key] = value

    def get_shape(self, key):
        return self.__d[key]



import logging

logger = logging.getLogger(__name__)


class BaseRowsStorage(object):

    # region abstract methods

    def _iter_rows(self):
        raise NotImplemented()

    def _get_rows_count(self):
        raise NotImplemented()

    # endregion

    # endregion

    # region base methods

    def __iter__(self):
        return self._iter_rows()

    def __len__(self):
        return self._get_rows_count()

    # endregion



import collections
import logging

logger = logging.getLogger(__name__)


class LabeledCollection:

    def __init__(self, uint_labeled_ids):
        """
        Args:
            uint_labeled_ids: collections.Iterable
                sequence of pairs (sample_row_id, uint_label).
        """
        assert(isinstance(uint_labeled_ids, collections.Iterable))

        self.__original_uint_labels = collections.OrderedDict(uint_labeled_ids)
        self.__assigned_uint_labels = {}

    def assign_uint_label(self, uint_label, sample_row_id):
        """
        Optionally applies the label.
        Applies label if the latter has not been provided before for sample_row_id
        """
        assert(isinstance(uint_label, int))
        assert(isinstance(sample_row_id, str))

        if sample_row_id not in self.__assigned_uint_labels:
            self.__assigned_uint_labels[sample_row_id] = uint_label

    def iter_non_duplicated_labeled_sample_row_ids(self):
        for sample_id, _ in self.__original_uint_labels.items():
            yield sample_id, self.__assigned_uint_labels[sample_id]



import tensorflow as tf
from collections import OrderedDict
from arenets.context.architectures.base.base import SingleInstanceNeuralNetwork


class FullyConnectedLayer(SingleInstanceNeuralNetwork):

    H_W = "W"
    H_b = "b"

    def __init__(self):
        super(FullyConnectedLayer, self).__init__()
        self.__hidden = OrderedDict()

    def init_logits_unscaled(self, context_embedding):

        with tf.name_scope("output"):
            logits = tf.compat.v1.nn.xw_plus_b(context_embedding,
                                               self.__hidden[self.H_W],
                                               self.__hidden[self.H_b],
                                               name="logits")

        return logits, tf.nn.dropout(logits, self.DropoutKeepProb)

    def init_logits_hidden_states(self):

        self.__hidden[self.H_W] = tf.compat.v1.get_variable(
            name=self.H_W,
            shape=[self.ContextEmbeddingSize, self.Config.ClassesCount],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer)

        self.__hidden[self.H_b] = tf.compat.v1.get_variable(
            name=self.H_b,
            shape=[self.Config.ClassesCount],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.BiasInitializer)

    # region public 'iter' methods

    def iter_hidden_parameters(self):
        for key, value in self.__hidden.items():
            yield key, value

    # endregion



import logging
import sys
import unittest

sys.path.append('../')

from tests.test_tf_ctx_feed import TestContextNetworkFeeding
from tests.tf_networks.supported import get_supported
from tests.labels import TestThreeLabelScaler

from arenets.arekit.common.labels.scaler import BaseLabelScaler
from arenets.shapes import NetworkInputShapes
from arenets.multi.configurations.att_self import AttSelfOverSentencesConfig
from arenets.multi.architectures.att_self import AttSelfOverSentences
from arenets.core.feeding.bags.bag import Bag
from arenets.core.feeding.batch.multi import MultiInstanceMiniBatch
from arenets.multi.configurations.max_pooling import MaxPoolingOverSentencesConfig
from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.sample import InputSample
from arenets.multi.architectures.max_pooling import MaxPoolingOverSentences


class TestMultiInstanceFeed(unittest.TestCase):

    @staticmethod
    def __create_minibatch(config, labels_scaler):
        assert(isinstance(config, DefaultNetworkConfig))
        assert(isinstance(labels_scaler, BaseLabelScaler))

        bags = []
        no_label = labels_scaler.get_no_label_instance()

        shapes = NetworkInputShapes(iter_pairs=[
            (NetworkInputShapes.FRAMES_PER_CONTEXT, config.FramesPerContext),
            (NetworkInputShapes.TERMS_PER_CONTEXT, config.TermsPerContext),
            (NetworkInputShapes.SYNONYMS_PER_CONTEXT, config.SynonymsPerContext)])

        empty_sample = InputSample.create_empty(shapes)

        for i in range(config.BagsPerMinibatch):
            bag = Bag(labels_scaler.label_to_uint(no_label))
            for j in range(config.BagSize):
                bag.add_sample(empty_sample)
            bags.append(bag)

        return MultiInstanceMiniBatch(bags=bags, batch_id=None)

    @staticmethod
    def multiinstances_supported(ctx_config, ctx_network):
        return [
            (MaxPoolingOverSentencesConfig(ctx_config), MaxPoolingOverSentences(ctx_network)),
            (AttSelfOverSentencesConfig(ctx_config), AttSelfOverSentences(ctx_network))
        ]

    def test(self):
        logger = logging.getLogger(__name__)
        logging.basicConfig(level=logging.INFO)

        labels_scaler = TestThreeLabelScaler()

        for ctx_config, ctx_network in get_supported():
            for config, network in self.multiinstances_supported(ctx_config, ctx_network):
                logger.info(type(network))
                logger.info('\t-> {}'.format(type(ctx_network)))
                TestContextNetworkFeeding.run_feeding(network=network,
                                                      network_config=config,
                                                      create_minibatch_func=self.__create_minibatch,
                                                      logger=logger,
                                                      labels_scaler=labels_scaler,
                                                      display_idp_values=False)


if __name__ == '__main__':
    unittest.main()



import tensorflow as tf

from arenets.context.configurations.base.base import DefaultNetworkConfig


class BaseMultiInstanceConfig(DefaultNetworkConfig):

    __contexts_per_opinion = 3
    __context_config = None

    def __init__(self, context_config):
        assert(isinstance(context_config, DefaultNetworkConfig))
        super(BaseMultiInstanceConfig, self).__init__()
        self.__context_config = context_config

        self.__context_parameters_fix()

    # region properties

    @property
    def BagsPerMinibatch(self):
        return super(BaseMultiInstanceConfig, self).BagsPerMinibatch

    @property
    def BatchSize(self):
        return self.BagsPerMinibatch

    @property
    def BagSize(self):
        return self.__contexts_per_opinion

    @property
    def ContextConfig(self):
        return self.__context_config

    @property
    def WeightInitializer(self):
        return tf.contrib.layers.xavier_initializer()

    @property
    def BaseInitializer(self):
        return tf.random_uniform_initializer(-0.1, 0.1)

    @property
    def ClassesCount(self):
        return self.__context_config.ClassesCount

    @property
    def LearningRate(self):
        return self.__context_config.LearningRate

    @property
    def TermsPerContext(self):
        return self.__context_config.TermsPerContext

    @property
    def EmbeddingDropoutKeepProb(self):
        return self.__context_config.EmbeddingDropoutKeepProb

    # endregion

    def set_pos_count(self, value):
        super(BaseMultiInstanceConfig, self).set_pos_count(value)
        self.__context_config.set_pos_count(value)

    def set_contexts_per_opinion(self, value):
        self.__contexts_per_opinion = value

    def set_term_embedding(self, embedding_matrix):
        super(BaseMultiInstanceConfig, self).set_term_embedding(embedding_matrix)
        self.__context_config.set_term_embedding(embedding_matrix)

    def set_class_weights(self, class_weights):
        super(BaseMultiInstanceConfig, self).set_class_weights(class_weights)
        self.__context_config.set_class_weights(class_weights)

    def modify_bags_per_minibatch(self, value):
        super(BaseMultiInstanceConfig, self).modify_bags_per_minibatch(value)
        self.__context_parameters_fix()

    def modify_classes_count(self, value):
        super(BaseMultiInstanceConfig, self).modify_classes_count(value)
        self.__context_config.modify_classes_count(value)

    def modify_dropout_keep_prob(self, value):
        super(BaseMultiInstanceConfig, self).modify_dropout_keep_prob(value)
        self.__context_config.modify_dropout_keep_prob(value)

    def modify_learning_rate(self, value):
        super(BaseMultiInstanceConfig, self).modify_learning_rate(value)
        self.__context_config.modify_learning_rate(value)

    def modify_use_class_weights(self, value):
        super(BaseMultiInstanceConfig, self).modify_use_class_weights(value)
        self.__context_config.modify_use_class_weights(value)

    def reinit_config_dependent_parameters(self):
        super(BaseMultiInstanceConfig, self).reinit_config_dependent_parameters()
        self.__context_config.reinit_config_dependent_parameters()

    def fix_context_parameters(self):
        self.__context_parameters_fix()

    def __context_parameters_fix(self):
        self.__context_config.modify_bag_size(1)
        self.__context_config.modify_bags_per_minibatch(self.BatchSize)

    def _internal_get_parameters(self):

        parameters = [
            ("mi:Batch (BagsPerMinibatch)", self.BagsPerMinibatch),
            ("mi:BagSize (ContextsPerOpinion)", self.BagSize),
            ("mi:LearningRate", self.LearningRate),
            ("mi:DropoutKeepProb", self.DropoutKeepProb),
            ("mi:Optimiser", self.Optimiser)
        ]

        parameters.extend(self.__context_config._internal_get_parameters())

        return parameters



import logging
import tensorflow as tf


logger = logging.getLogger(__name__)


def get_k_layer_logits(g, W, b, dropout_keep_prob=None, activations=None):
    assert(isinstance(W, list))
    assert(isinstance(b, list))
    assert(isinstance(activations, list) or activations is None)

    if activations is None:
        activations = [None] * (len(W) + 1)

    assert(len(W) == len(b) == len(activations) - 1)

    def __activate(tensor, activation):
        return activation(tensor) if activation is not None else tensor

    res = g

    for i in range(len(W)):

        logger.debug("LOG: r_shape={}".format(res.shape))
        logger.debug("LOG: i={}".format(i))
        logger.debug("LOG: W[i]_shape={}".format(W[i].shape))
        logger.debug("LOG: b[i]_shape={}".format(b[i].shape))

        res = __activate(res, activations[i])
        res = tf.matmul(res, W[i]) + b[i]

        if dropout_keep_prob is None:
            continue

        res = tf.nn.dropout(res, dropout_keep_prob)

    return __activate(res, activations[-1])


def get_k_layer_pair_logits(g, W, b, dropout_keep_prob, activations=None):
    assert(dropout_keep_prob is not None)

    result = get_k_layer_logits(g, W, b, activations=activations)
    result_dropout = get_k_layer_logits(g, W, b,
                                        dropout_keep_prob=dropout_keep_prob,
                                        activations=activations)

    return result, result_dropout



import logging

from arenets.arekit.contrib.utils.np_utils.npz_utils import NpzRepositoryUtils

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class NpzEmbeddingHelper:

    @staticmethod
    def load_embedding(source):
        embedding = NpzRepositoryUtils.load(source)
        logger.info("Embedding read [size={size}]: {filepath}".format(size=embedding.shape,
                                                                      filepath=source))
        return embedding
    
    @staticmethod
    def save_embedding(data, target):
        NpzRepositoryUtils.save(data=data, target=target)
        logger.info("Saving embedding [size={shape}]: {filepath}".format(shape=data.shape,
                                                                         filepath=target))



from collections import OrderedDict


class InputSampleBase(object):
    """
    Description of a single sample (context) of a model
    """

    def __init__(self, shift_index_dbg, input_sample_id, values):
        assert(isinstance(shift_index_dbg, int))
        assert(isinstance(input_sample_id, str) or isinstance(input_sample_id, int))
        assert(isinstance(values, list))
        self._shift_index_dbg = shift_index_dbg
        self.__input_sample_id = str(input_sample_id) if isinstance(input_sample_id, int) else input_sample_id
        self.__values = OrderedDict(values)

    # region properties

    @property
    def ID(self):
        return self.__input_sample_id

    # endregion

    @staticmethod
    def _check_ends_could_be_fitted_in_window(actual_dist, window):
        return actual_dist < window

    def __iter__(self):
        for key, value in self.__values.items():
            yield key, value



"""
(c) Ilya Ivanov
github: https://github.com/ilivans
code: https://github.com/ilivans/tf-rnn-attention
"""

import tensorflow as tf


def self_attention_by_z_yang(inputs, attention_size, time_major=False):
    """
    Attention mechanism layer which reduces RNN/Bi-RNN outputs with Attention vector.
    The idea was proposed in the article by Z. Yang et al., "Hierarchical Attention Networks
     for Document Classification", 2016: http://www.aclweb.org/anthology/N16-1174.
    Variables notation is also inherited from the article

    Args:
        inputs: The Attention inputs.
            Matches outputs of RNN/Bi-RNN layer (not final state):
                In case of RNN, this must be RNN outputs `Tensor`:
                    If time_major == False (default), this must be a tensor of shape:
                        `[batch_size, max_time, cell.output_size]`.
                    If time_major == True, this must be a tensor of shape:
                        `[max_time, batch_size, cell.output_size]`.
                In case of Bidirectional RNN, this must be a tuple (outputs_fw, outputs_bw) containing the forward and
                the backward RNN outputs `Tensor`.
                    If time_major == False (default),
                        outputs_fw is a `Tensor` shaped:
                        `[batch_size, max_time, cell_fw.output_size]`
                        and outputs_bw is a `Tensor` shaped:
                        `[batch_size, max_time, cell_bw.output_size]`.
                    If time_major == True,
                        outputs_fw is a `Tensor` shaped:
                        `[max_time, batch_size, cell_fw.output_size]`
                        and outputs_bw is a `Tensor` shaped:
                        `[max_time, batch_size, cell_bw.output_size]`.
        attention_size: Linear size of the Attention weights.
        time_major: The shape format of the `inputs` Tensors.
            If true, these `Tensors` must be shaped `[max_time, batch_size, depth]`.
            If false, these `Tensors` must be shaped `[batch_size, max_time, depth]`.
            Using `time_major = True` is a bit more efficient because it avoids
            transposes at the beginning and end of the RNN calculation.  However,
            most TensorFlow data is batch-major, so by default this function
            accepts input and emits output in batch-major form.
    Returns:
        alphas: Tensor
            tensor of shape (B, T), which corresponds to weights of the particular time-series item.
    """

    if isinstance(inputs, tuple):
        # In case of Bi-RNN, concatenate the forward and the backward RNN outputs.
        inputs = tf.concat(inputs, 2)

    if time_major:
        # (T,B,D) => (B,T,D)
        inputs = tf.array_ops.transpose(inputs, [1, 0, 2])

    hidden_size = inputs.shape[2].value  # D value - hidden size of the RNN layer

    initializer = tf.random_normal_initializer(stddev=0.1)

    # Trainable parameters
    w_omega = tf.compat.v1.get_variable(name="w_omega", shape=[hidden_size, attention_size], initializer=initializer)
    b_omega = tf.compat.v1.get_variable(name="b_omega", shape=[attention_size], initializer=initializer)
    u_omega = tf.compat.v1.get_variable(name="u_omega", shape=[attention_size], initializer=initializer)

    with tf.name_scope('v'):
        # Applying fully connected layer with non-linear activation to each of the B*T timestamps;
        #  the shape of `v` is (B,T,D)*(D,A)=(B,T,A), where A=attention_size
        v = tf.tanh(tf.tensordot(inputs, w_omega, axes=1) + b_omega)

    # For each of the timestamps its vector of size A from `v` is reduced with `u` vector
    vu = tf.tensordot(v, u_omega, axes=1, name='vu')  # (B,T) shape
    alphas = tf.nn.softmax(vu, name='alphas')  # (B,T) shape

    return alphas


def calculate_sequential_attentive_weights_by_z_yang(inputs, alphas):
    """Attention mechanism layer which reduces RNN/Bi-RNN outputs with Attention vector.
       The idea was proposed in the article by Z. Yang et al., "Hierarchical Attention Networks
       for Document Classification", 2016: http://www.aclweb.org/anthology/N16-1174.
       Variables notation is also inherited from the article

        Proposes to reduce the time-series parameter
        inputs: Tensor
            tensor of shape (B, T, D), where
                B -- batch, T -- time-series (context terms), D -- data or hidden embedding.
        alphas: Tensor
            tensor of shape (B, T), which corresponds to weights of the particular time-series item.
    """

    # Output of (Bi-)RNN is reduced with attention vector; the result has (B,D) shape
    output = tf.reduce_sum(inputs * tf.expand_dims(alphas, -1), 1)

    return output



from arenets.arekit.common.pipeline.context import PipelineContext
from arenets.core.feeding.batch.base import MiniBatch
from arenets.core.pipeline.item_base import EpochHandlingPipelineItem


class EpochLabelsPredictorPipelineItem(EpochHandlingPipelineItem):

    KEY = "uint_labels"

    def apply_core(self, input_data, pipeline_ctx):
        assert(isinstance(input_data, MiniBatch))
        assert(isinstance(pipeline_ctx, PipelineContext))
        feed_dict = self._context.create_feed_dict(minibatch=input_data, data_type=self._data_type)
        uint_labels = self._context.Session.run(self._context.Network.Labels, feed_dict=feed_dict)
        pipeline_ctx.update(param=self.KEY, value=uint_labels)



import tensorflow as tf
from collections import OrderedDict
from arenets.multi.architectures.base.base_single_mlp import BaseMultiInstanceSingleMLP


class MaxPoolingOverSentences(BaseMultiInstanceSingleMLP):
    """
    Title: Relation Extraction with Multi-instance Multi-label Convolutional Neural Networks
    Authors: Xiaotian Jiang, Quan Wang, Peng Li, Bin Wang
    Paper: https://pdfs.semanticscholar.org/8731/369a707046f3f8dd463d1fd107de31d40a24.pdf
    """

    def __init__(self, context_network):
        super(MaxPoolingOverSentences, self).__init__(context_network)
        self.__hidden = OrderedDict()

    @property
    def EmbeddingSize(self):
        return self.ContextNetwork.ContextEmbeddingSize

    def init_multiinstance_embedding(self, context_outputs):
        """
        context_outputs: [batches, sentences, embedding]
        """
        return self.__contexts_max_pooling(context_outputs=context_outputs,
                                           contexts_per_opinion=self.ContextsPerOpinion)  # [batches, max_pooling]

    @staticmethod
    def __contexts_max_pooling(context_outputs, contexts_per_opinion):
        context_outputs = tf.expand_dims(context_outputs, 0)     # [1, batches, sentences, embedding]
        context_outputs = tf.nn.max_pool2d(
            input=context_outputs,
            ksize=[1, 1, contexts_per_opinion, 1],
            strides=[1, 1, contexts_per_opinion, 1],
            padding='VALID',
            data_format="NHWC")
        return tf.squeeze(context_outputs)                       # [batches, max_pooling]



from arenets.sample import InputSample


class Bag:
    """
    Bag is a list of samples (of related contexts)
    and an attitude label
    """

    def __init__(self, uint_label):
        assert(isinstance(uint_label, int) or uint_label is None)
        self.__samples = []
        self.__uint_label = uint_label

    # region properties

    @property
    def UintBagLabel(self):
        return self.__uint_label

    @property
    def Samples(self):
        return self.__samples

    # endregion

    # region public methods

    def add_sample(self, sample):
        assert(isinstance(sample, InputSample))
        self.__samples.append(sample)

    # endregion

    # region overriden methods

    def __len__(self):
        return len(self.__samples)

    def __iter__(self):
        for sample in self.__samples:
            yield sample

    # endregion



import glob
import os
import shutil


def rm_dir_contents(dir_path, logger):
    contents = glob.glob(dir_path)
    for f in contents:
        logger.info("Removing old file/dir: {}".format(f))
        if os.path.isfile(f):
            os.remove(f)
        else:
            shutil.rmtree(f, ignore_errors=True)



from arenets.arekit.common.pipeline.context import PipelineContext
from arenets.core.idhv_collection import NetworkInputDependentVariables
from arenets.core.pipeline.item_base import EpochHandlingPipelineItem


class MinibatchHiddenFetcherPipelineItem(EpochHandlingPipelineItem):

    def __init__(self):
        super(MinibatchHiddenFetcherPipelineItem, self).__init__()
        self.__idh_names = None
        self.__idh_tensors = None
        self.__input_dependent_params = None

    @property
    def InputDependentParams(self):
        return self.__input_dependent_params

    def before_epoch(self, model_context, data_type):
        super(MinibatchHiddenFetcherPipelineItem, self).before_epoch(
            model_context=model_context, data_type=data_type)

        self.__idh_names, self.__idh_tensors = map(
            list, zip(*self._context.Network.iter_input_dependent_hidden_parameters()))

        self.__input_dependent_params = NetworkInputDependentVariables()

    def apply_core(self, input_data, pipeline_ctx):
        assert(isinstance(pipeline_ctx, PipelineContext))
        minibatch = input_data
        feed_dict = self._context.create_feed_dict(minibatch=minibatch, data_type=self._data_type)
        idh_values = self._context.Session.run(self.__idh_tensors, feed_dict=feed_dict)

        if not (len(self.__idh_names) > 0 and len(idh_values) > 0):
            return

        self.__input_dependent_params.add_input_dependent_values(
            names_list=self.__idh_names,
            tensor_values_list=idh_values,
            text_opinion_ids=[sample.ID for sample in minibatch.iter_by_samples()],
            bags_per_minibatch=self._context.Config.BagsPerMinibatch,
            bag_size=self._context.Config.BagSize)



from arenets.attention.architectures.self_z_yang import self_attention_by_z_yang, \
    calculate_sequential_attentive_weights_by_z_yang
from arenets.context.architectures.base.att_self_bilstm_base import AttentionSelfBiLSTMBase
from arenets.context.configurations.att_self_z_yang_bilstm import AttentionSelfZYangBiLSTMConfig


class AttentionSelfZYangBiLSTM(AttentionSelfBiLSTMBase):

    def get_attention_output_with_alphas(self, rnn_outputs):
        assert(isinstance(self.Config, AttentionSelfZYangBiLSTMConfig))
        alphas = self_attention_by_z_yang(inputs=rnn_outputs, attention_size=self.Config.AttentionSize)
        attentive_output = calculate_sequential_attentive_weights_by_z_yang(inputs=rnn_outputs, alphas=alphas)
        return attentive_output, alphas



import collections
import logging

from arenets.arekit.common.data_type import DataType
from arenets.arekit.common.pipeline.base import BasePipeline
from arenets.arekit.common.utils import progress_bar_defined
from arenets.arekit.model.base import BaseModel

from arenets.core.cancellation import OperationCancellation
from arenets.core.feeding.bags.collection.factory import create_batch_by_bags_group
from arenets.core.model_ctx import TensorflowModelContext
from arenets.core.params import NeuralNetworkModelParams
from arenets.core.pipeline.item_base import EpochHandlingPipelineItem

from arenets.tf_helpers.nn_states import TensorflowNetworkStatesProvider

logger = logging.getLogger(__name__)


class BaseTensorflowModel(BaseModel):

    SaveTensorflowModelStateOnFit = True

    def __init__(self, context, callbacks,
                 predict_pipeline=None,
                 fit_pipeline=None):
        assert(isinstance(context, TensorflowModelContext))
        assert(isinstance(callbacks, list))
        assert(isinstance(predict_pipeline, list))
        assert(isinstance(fit_pipeline, list))
        super(BaseTensorflowModel, self).__init__()

        self.__context = context
        self.__callbacks = callbacks
        self.__predict_pipeline = predict_pipeline
        self.__fit_pipeline = fit_pipeline
        self.__states_provider = TensorflowNetworkStatesProvider()

    @property
    def Context(self):
        return self.__context

    # region private methods

    def __callback_do(self, call_func):
        for callback in self.__callbacks:
            call_func(callback)

    @staticmethod
    def __handle_batches_iter(batches_iter, total, prefix, unit='mbs'):
        """ Do wrapping progress notification.
        """
        assert(isinstance(batches_iter, collections.Iterable))
        assert(isinstance(unit, str))
        assert(isinstance(prefix, str))
        desc = "{prefix}".format(prefix=prefix)
        return progress_bar_defined(iterable=batches_iter, unit=unit, total=total, desc=desc)

    def __run_epoch_pipeline(self, data_type, pipeline_items, prefix):
        assert(isinstance(pipeline_items, list))
        assert(isinstance(prefix, str))

        bags_per_group = self.__context.Config.BagsPerMinibatch
        bags_collection = self.__context.get_bags_collection(data_type)
        minibatches_count = bags_collection.get_groups_count(bags_per_group)

        logger.info("Minibatches passing per epoch count: ~{} "
                    "(Might be greater or equal, as the last "
                    "bag is expanded)".format(minibatches_count))

        groups_it = self.__handle_batches_iter(
            batches_iter=bags_collection.iter_by_groups(bags_per_group=bags_per_group),
            total=minibatches_count,
            prefix=prefix)

        for item in pipeline_items:
            assert(isinstance(item, EpochHandlingPipelineItem))
            item.before_epoch(model_context=self.__context,
                              data_type=data_type)

        pipeline = BasePipeline(pipeline_items)

        for bags_group in groups_it:
            assert(isinstance(bags_group, list))

            # Composing minibatch from bags group.
            minibatch = create_batch_by_bags_group(
                bags_collection_type=self.__context.BagsCollectionType,
                bags_group=bags_group)

            pipeline.run(minibatch)

    def __try_load_state(self):
        if self.__context.IO.IsPretrainedStateProvided:
            self.__states_provider.load_model(sess=self.__context.Session,
                                              path_tf_prefix=self.__context.IO.get_model_source_path_tf_prefix())

    def __fit(self, epochs_count, data_type):
        assert(isinstance(epochs_count, int))
        assert(isinstance(data_type, DataType))
        assert(self.__context.Session is not None)

        operation_cancel = OperationCancellation()
        bags_collection = self.__context.get_bags_collection(data_type)

        self.__callback_do(lambda callback: callback.on_fit_started(operation_cancel))

        for epoch_index in range(epochs_count):

            if operation_cancel.IsCancelled:
                break

            bags_collection.shuffle()

            self.__run_epoch_pipeline(pipeline_items=self.__fit_pipeline,
                                      data_type=data_type,
                                      prefix="Training [{data_type}]".format(data_type=data_type))

            self.__callback_do(lambda callback: callback.on_epoch_finished(
                pipeline=self.__fit_pipeline,
                operation_cancel=operation_cancel))

            if BaseTensorflowModel.SaveTensorflowModelStateOnFit:
                self.__states_provider.save_model(sess=self.__context.Session,
                                                  path_tf_prefix=self.__context.IO.get_model_target_path_tf_prefix())

    # endregion

    def fit(self, model_params, seed, data_type=DataType.Train):
        assert(isinstance(model_params, NeuralNetworkModelParams))
        self.__context.Network.compile(self.__context.Config, reset_graph=True, graph_seed=seed)
        self.__context.set_optimiser()
        self.__context.initialize_session()
        self.__try_load_state()
        self.__fit(epochs_count=model_params.EpochsCount, data_type=data_type)

        self.__context.dispose_session()

    def predict(self, data_type=DataType.Test, do_compile=False, graph_seed=0):

        # Optionally perform network compilation
        if do_compile:
            self.__context.Network.compile(config=self.__context.Config, reset_graph=True, graph_seed=graph_seed)

        self.__context.initialize_session()
        self.__try_load_state()
        self.__run_epoch_pipeline(pipeline_items=self.__predict_pipeline,
                                  data_type=data_type,
                                  prefix="Predict [{dtype}]".format(dtype=data_type))
        self.__callback_do(lambda callback: callback.on_predict_finished(self.__predict_pipeline))
